from typing import Literal

from pydantic import BaseModel, Field

from .base import Base, BaseRelationships, Relationship


class PlayerQueryConditions(BaseModel):
    """Conditions for the PlayerQuery model."""

    score: float
    score_type: Literal["score", "pow", "multiplier"] = Field(alias="scoreType")
    types: list[str | None]

    model_config = {
        "populate_by_name": True,
    }


class PlayerQueryAttributes(BaseModel):
    """Attributes for the PlayerQuery model."""

    conditions: PlayerQueryConditions
    created_at: str = Field(alias="createdAt")
    query_name: str = Field(alias="queryName")
    updated_at: str = Field(alias="updatedAt")

    model_config = {
        "populate_by_name": True,
    }


class PlayerQueryRelationships(BaseRelationships):
    """Relationships for the PlayerQuery model."""

    organization: Relationship
    user: Relationship


class PlayerQuery(Base):
    """PlayerQuery model representing a query for players in Battlemetrics."""

    type: str = "playerQuery"
    attributes: PlayerQueryAttributes
    relationships: PlayerQueryRelationships


class PlayerQueryResultAttributes(BaseModel):
    """Attributes for the PlayerQueryResult model."""

    score: int


class PlayerQueryResultRelationships(BaseRelationships):
    """Relationships for the PlayerQueryResult model."""

    player: Relationship


class PlayerQueryResult(Base):
    """PlayerQueryResult model representing a result from a player query."""

    type: str = "playerQueryResult"
    attributes: PlayerQueryResultAttributes
    relationships: PlayerQueryResultRelationships
    relationships: PlayerQueryResultRelationships



from typing import Literal

from pydantic import BaseModel, Field

from .base import Base, BaseRelationships, Relationship
from .server import ServerData


class BanListAttributes(BaseModel):
    """Attributes for the BanList model."""

    action: Literal["none", "log", "kick"] | None
    default_auto_add_enabled: bool = Field(alias="defaultAutoAddEnabled")
    default_identifiers: list[str] = Field(
        default_factory=list,
        alias="defaultIdentifiers",
    )
    default_native_enabled: bool | None = Field(
        default=None,
        alias="defaultNativeEnabled",
    )
    default_reasons: list[str] = Field(default_factory=list, alias="defaultReasons")
    name: str
    native_ban_perm_max_expires: int | None = Field(
        default=None,
        alias="nativeBanPermMaxExpires",
    )
    native_ban_ttl: int | None = Field(default=None, alias="nativeBanTTL")
    native_ban_temp_max_expires: int | None = Field(
        default=None,
        alias="nativeBanTempMaxExpires",
    )
    perm_create: bool = Field(alias="permCreate")
    perm_delete: bool = Field(alias="permDelete")
    perm_manage: bool = Field(alias="permManage")
    perm_update: bool = Field(alias="permUpdate")

    model_config = {
        "populate_by_name": True,
    }


class BanListRelationships(BaseRelationships):
    """Relationships for the BanList model."""

    organization: Relationship
    owner: Relationship
    servers: list[ServerData] | None = None


class BanList(Base):
    """Represents a ban list in Battlemetrics."""

    type: str = "banList"
    attributes: BanListAttributes
    relationships: BanListRelationships


class BanListExemptionAttributes(BaseModel):
    """Attributes for the BanListExemption model."""

    reason: str


class BanListExemptionRelationships(BaseRelationships):
    """Relationships for the BanListExemption model."""

    ban: Relationship
    ban_list: Relationship = Field(alias="banList")


class BanListExemption(Base):
    """Represents an exemption from a ban list."""

    type: str = "banExemption"
    attributes: BanListExemptionAttributes
    relationships: BanListExemptionRelationships
    relationships: BanListExemptionRelationships


class BanListInviteAttributes(BaseModel):
    """Attributes for the BanListInvite model."""

    limit: int | None
    perm_create: bool = Field(alias="permCreate")
    perm_delete: bool = Field(alias="permDelete")
    perm_manage: bool = Field(alias="permManage")
    perm_update: bool = Field(alias="permUpdate")
    uses: int

    model_config = {
        "populate_by_name": True,
    }


class BanListInviteRelationships(BaseRelationships):
    """Relationships for the BanListInvite model."""

    ban_list: Relationship = Field(alias="banList")
    organization: Relationship
    user: Relationship


class BanListInvite(Base):
    """Represents an invite to a ban list."""

    type: str = "banListInvite"
    attributes: BaseModel
    relationships: BaseRelationships



from .ban import *
from .banlist import *
from .base import *
from .flag import *
from .game import *
from .misc import *
from .note import *
from .organization import *
from .player import *
from .query import *
from .server import *
from .session import *
from .user import *



from typing import Literal

from pydantic import BaseModel, Field

from .base import Base, BaseRelationships, Relationship


class GameMetadata(BaseModel):
    """Metadata specific to the Game model.

    Fields may be partially present depending on the game; treat them as optional.
    """

    appid: float | None = None
    gamedir: str | None = None
    no_player_list: bool | None = Field(default=None, alias="noPlayerList")

    model_config = {
        "extra": "allow",
        "populate_by_name": True,
    }


class GameAttributes(BaseModel):
    """Attributes specific to the Game model."""

    max_players_24h: float = Field(alias="maxPlayers24H")
    max_players_30d: float = Field(alias="maxPlayers30D")
    max_players_7d: float = Field(alias="maxPlayers7D")
    min_players_24h: float = Field(alias="minPlayers24H")
    min_players_30d: float = Field(alias="minPlayers30D")
    min_players_7d: float = Field(alias="minPlayers7D")
    name: str
    players: int
    players_by_country: dict[str, int] = Field(alias="playersByCountry")
    servers: int
    servers_by_country: dict[str, int] = Field(alias="serversByCountry")
    metadata: GameMetadata | None = None

    model_config = {
        "populate_by_name": True,
    }


class Game(Base):
    """Game model representing a game in Battlemetrics."""

    type: str = "game"
    attributes: GameAttributes


class GameFeaturesAttributes(BaseModel):
    """Attributes for game features."""

    display: str
    feature_type: Literal["select", "boolean", "range", "dateRange"] = Field(
        alias="featureType",
    )
    type_options: str = Field(alias="typeOptions")

    model_config = {
        "populate_by_name": True,
    }


class GameFeaturesRelationships(BaseRelationships):
    """Relationships for game features."""

    game: Relationship


class GameFeatures(Base):
    """Game features model representing features of a game in Battlemetrics."""

    type: str = "gameFeature"
    attributes: GameFeaturesAttributes
    relationships: GameFeaturesRelationships


class GameFeatureOptionsAttributes(BaseModel):
    """Attributes for game feature options."""

    count: int
    display: str
    players: int
    updated_at: str = Field(alias="updatedAt")

    model_config = {
        "populate_by_name": True,
    }


class GameFeatureOptionsRelationships(BaseRelationships):
    """Relationships for game feature options."""

    game_feature: Relationship = Field(alias="gameFeature")


class GameFeatureOptions(Base):
    """Game feature options model representing options for a game feature in Battlemetrics."""

    type: str = "gameFeatureOption"
    attributes: GameFeatureOptionsAttributes
    relationships: GameFeatureOptionsRelationships



from pydantic import BaseModel, Field

from .base import Base, BaseRelationships, Relationship


class OrganizationAttributes(BaseModel):
    """Attributes specific to the Organization model."""

    active: bool
    ban_template: str = Field(alias="banTemplate")
    consent_api_keys_required: bool = Field(alias="consentAPIKeysRequired")
    consent_geo_ip_required: bool = Field(alias="consentGeoIPRequired")
    consent_organizations_required: bool = Field(alias="consentOrganizationsRequired")
    data_sharing_enabled: bool = Field(alias="dataSharingEnabled")
    discoverable: bool
    discoverable_rank: int | None = Field(default=None, alias="discoverableRank")
    locale: str | None = None
    mfa_required: bool = Field(alias="mfaRequired")
    name: str
    plan: str | None = None
    tz: str | None = None

    model_config = {
        "populate_by_name": True,
    }


class OrganizationRelationships(BaseRelationships):
    """Relationships for the Organization model."""

    ban_lists: Relationship = Field(alias="banLists")
    default_ban_list: Relationship | None = Field(default=None, alias="defaultBanList")
    games: Relationship
    owner: Relationship
    servers: Relationship


class Organization(Base):
    """Organization model representing an organization in Battlemetrics."""

    type: str = "organization"
    attributes: OrganizationAttributes
    relationships: OrganizationRelationships


class OrganizationFriendAttributes(BaseModel):
    """Attributes for the OrganizationFriend model."""

    accepted: bool
    identifiers: list[str] | None
    notes: bool
    reciprocated: bool


class OrganizationFriendRelationships(BaseRelationships):
    """Relationships for the OrganizationFriend model."""

    flags_shared: Relationship = Field(alias="flagsShared")
    flags_used: Relationship = Field(alias="flagsUsed")
    friend: Relationship
    organization: Relationship


class OrganizationFriend(Base):
    """Represents a friendship between two organizations."""

    type: str = "organizationFriend"
    attributes: OrganizationFriendAttributes
    relationships: OrganizationFriendRelationships


class OrganizationStatsAttributes(BaseModel):
    """Attributes for the Organization Stats model."""

    game_ranks: dict[str, int] = Field(alias="gameRanks")
    identifiers: float
    unique_players: float = Field(alias="uniquePlayers")

    model_config = {
        "populate_by_name": True,
    }


class OrganizationStatsRelationships(BaseRelationships):
    """Relationships for the Organization Stats model."""

    organization: Relationship


class OrganizationStats(Base):
    """Represents the stats of an organization."""

    type: str = "organizationStats"
    attributes: OrganizationStatsAttributes
    relationships: OrganizationStatsRelationships



from pydantic import BaseModel, Field

from .base import Base, BaseRelationships, Relationship


class SessionMetadata(BaseModel):
    """Metadata for the Session model."""

    key: str
    private: bool
    value: str | None = None


class SessionAttributes(BaseModel):
    """Attributes for the Session model."""

    first_time: bool = Field(alias="firstTime")
    metadata: list[SessionMetadata] | None = None
    name: str
    private: bool
    start: str
    stop: str | None = None

    model_config = {
        "populate_by_name": True,
    }


class IdentifierRelationship(BaseModel):
    """Relationship to an identifier."""

    id: str
    type: str = "identifier"


class SessionRelationships(BaseRelationships):
    """Relationships for the Session model."""

    identifiers: list[IdentifierRelationship] | None = None
    organization: Relationship | None = None
    player: Relationship | None = None
    server: Relationship | None = None


class Session(Base):
    """Session model representing a game play session on a server."""

    type: str = "session"
    attributes: SessionAttributes
    relationships: SessionRelationships



from pydantic import BaseModel, Field

from .base import Base, BaseRelationships, Relationship


class NoteAttributes(BaseModel):
    """Attributes for the Note model."""

    clearance_level: int = Field(alias="clearanceLevel")
    created_at: str = Field(alias="createdAt")
    expires_at: str | None = Field(default=None, alias="expiresAt")
    note: str
    shared: bool

    model_config = {
        "populate_by_name": True,
    }


class NoteRelationships(BaseRelationships):
    """Relationships for the Note model."""

    organization: Relationship
    player: Relationship
    user: Relationship


class Note(Base):
    """Note model representing a note in Battlemetrics."""

    type: str = "playerNote"
    attributes: NoteAttributes
    relationships: NoteRelationships



from typing import Any

from aiohttp import ClientResponse

__all__ = ("BMException", "HTTPException", "Unauthorized", "Forbidden", "NotFound")


class BMException(Exception):
    """Base exception class for all BattleMetrics errors."""


class HTTPException(BMException):
    """Exception that's raised when an HTTP request operation fails.

    Attributes
    ----------
    response: :class:`aiohttp.ClientResponse`
        The response of the failed HTTP request. This is an
        instance of :class:`aiohttp.ClientResponse`. In some cases
        this could also be a :class:`requests.Response`.

    text: :class:`str`
        The text of the error. Could be an empty string.
    status: :class:`int`
        The status code of the HTTP request.
    code: :class:`int`
        The Discord specific error code for the failure.
    """

    def __init__(
        self,
        response: ClientResponse,
        message: str | dict[str, Any] | None,
    ) -> None:
        self.response: ClientResponse = response
        self.status: int = response.status  # type: ignore[reportAccessAttributeIssue]
        self.code: int = 0
        self.text: str = ""
        self.title: str = ""
        if isinstance(message, dict):
            error = message["errors"][0]
            self.code = (
                int(error.get("status", ""))
                if str(error.get("status", "")).isdigit()
                else 0
            )
            self.text = error.get("title", "")
            self.title = error.get("title", "")
        else:
            self.text = str(message) if message else ""
            self.code = 0
        fmt = "{0.status} {0.reason}"
        if len(self.text):
            fmt += ": {1}"

        super().__init__(fmt.format(self.response, self.text))


class Unauthorized(HTTPException):
    """Exception that's raised for when status code 401 occurs."""


class Forbidden(HTTPException):
    """Exception that's raised for when status code 403 occurs."""


class NotFound(HTTPException):
    """Exception that's raised for when status code 404 occurs."""



"""User related models for the Battlemetrics API."""

from pydantic import BaseModel

from .base import Base, BaseRelationships, Relationship


class UserAttributes(BaseModel):
    """Attributes for the User model."""

    id: str
    nickname: str | None = None


class UserRelationships(BaseRelationships):
    """Relationships for the User model."""

    organizations: list[Relationship]


class User(Base):
    """User model representing a user in Battlemetrics."""

    type: str = "user"
    attributes: UserAttributes
    relationships: UserRelationships



from typing import Any, Literal

from pydantic import BaseModel, Field

from .base import Base, BaseRelationships, Relationship


class ServerMetadata(BaseModel):
    """Metadata specific to the Server model."""

    beta_source_protocol: bool | None = Field(default=None, alias="betaSourceProtocol")
    client_perf: bool | None = Field(default=None, alias="clientPerf")
    connection_type: Literal["source", "ws"] | None = Field(
        default=None,
        alias="connectionType",
    )
    disable_locked: bool | None = Field(default=None, alias="disableLocked")
    disabled_reason: str | None = Field(default=None, alias="disabledReason")
    has_source_mod: bool | None = Field(default=None, alias="hasSourceMod")
    hll_get_player_info: bool | None = Field(default=None, alias="hllGetPlayerInfo")
    hll_player_list_interval: int | None = Field(
        default=None,
        alias="hllPlayerListInterval",
    )
    log_secret: str | None = Field(default=None, alias="logSecret")
    private_player_sessions: bool | None = Field(
        default=None,
        alias="privatePlayerSessions",
    )
    rcon_ip: str | None = Field(default=None, alias="rconIP")
    reserved_slot_kick_reason: str | None = Field(
        default=None,
        alias="reservedSlotKickReason",
    )
    reserved_slots: int | None = Field(default=None, alias="reservedSlots")
    reserved_slots_kick_last_to_join: bool | None = Field(
        default=None,
        alias="reservedSlotsKickLastToJoin",
    )
    status_interval: int | None = Field(default=None, alias="statusInterval")
    use_connection_pool: bool | None = Field(default=None, alias="useConnectionPool")
    use_get_chat: bool | None = Field(default=None, alias="useGetChat")
    username: str | None = None

    model_config = {
        "populate_by_name": True,
    }


class ServerAttributes(BaseModel):
    """Attributes specific to the Server model."""

    address: str | None = None
    country: str
    created_at: str = Field(alias="createdAt")
    details: dict[str, Any] | None = None
    id: str
    ip: str
    location: list[float]
    max_players: int = Field(alias="maxPlayers")
    metadata: ServerMetadata | None = None
    name: str
    players: int
    port: int
    port_query: int = Field(alias="portQuery")
    private: bool
    query_status: str | None = Field(default=None, alias="queryStatus")
    rank: int | None = None
    rcon_active: bool | None = Field(default=None, alias="rconActive")
    rcon_disconnected: str | None = Field(default=None, alias="rconDisconnected")
    rcon_last_connected: str | None = Field(default=None, alias="rconLastConnected")
    rcon_status: str | None = Field(default=None, alias="rconStatus")
    status: str
    updated_at: str = Field(alias="updatedAt")

    model_config = {
        "populate_by_name": True,
    }


class ServerGroupRelationship(Relationship):
    """Relationship for the ServerGroup model."""

    class ServerGroupMeta(BaseModel):
        """Metadata for the ServerGroup model."""

        leader: bool

    meta: ServerGroupMeta | None = None


class ServerRelationships(BaseRelationships):
    """Relationships for the Server model."""

    default_ban_list: Relationship | None = Field(default=None, alias="defaultBanList")
    game: Relationship | None = None
    organization: Relationship | None = None
    server_group: ServerGroupRelationship | None = Field(
        default=None,
        alias="serverGroup",
    )


class ServerMeta(BaseModel):
    """Metadata for the Server model."""

    action: Literal["none", "log", "kick"] | None
    default_native_enabled: bool | None = Field(
        default=None,
        alias="defaultNativeEnabled",
    )
    native_ban_perm_max_expires: int | None = Field(
        default=None,
        alias="nativeBanPermMaxExpires",
    )
    native_ban_ttl: int | None = Field(default=None, alias="nativeBanTTL")
    native_ban_temp_max_expires: int | None = Field(
        default=None,
        alias="nativeBanTempMaxExpires",
    )

    model_config = {
        "populate_by_name": True,
    }


class ServerData(Base):
    """Server data model representing a server in Battlemetrics."""

    type: str = "server"
    meta: ServerMeta | None = None


class Server(Base):
    """Server model representing a server in Battlemetrics."""

    type: str = "server"
    attributes: ServerAttributes
    relationships: ServerRelationships


class ReservedSlotAttributes(BaseModel):
    """Attributes for the ReservedSlot model."""

    created_at: str = Field(alias="createdAt")
    expires: str | None = None
    identifiers: list[str]

    model_config = {
        "populate_by_name": True,
    }


class ReservedSlotRelationships(BaseRelationships):
    """Relationships for the ReservedSlot model."""

    organization: Relationship
    player: Relationship
    servers: Relationship
    user: Relationship | None = None


class ReservedSlot(Base):
    """ReservedSlot model representing a reserved slot in Battlemetrics."""

    type: str = "reservedSlot"
    attributes: ReservedSlotAttributes
    meta: dict[str, Any]



from __future__ import annotations

import asyncio
import warnings
from typing import TYPE_CHECKING, Any, Literal

from battlemetrics.http import HTTPClient
from battlemetrics.models.ban import Ban, NativeBan
from battlemetrics.models.banlist import BanList, BanListExemption, BanListInvite
from battlemetrics.models.flag import FlagPlayer, PlayerFlag
from battlemetrics.models.game import Game, GameFeatureOptions, GameFeatures
from battlemetrics.models.misc import CommandsActivity, Coplay, Leaderboard
from battlemetrics.models.note import Note
from battlemetrics.models.organization import (
    Organization,
    OrganizationFriend,
    OrganizationStats,
)
from battlemetrics.models.player import (
    Player,
    QuickMatchIdentifier,
    RelatedPlayerIdentifier,
)
from battlemetrics.models.query import PlayerQuery, PlayerQueryResult
from battlemetrics.models.server import ReservedSlot, Server
from battlemetrics.models.session import Session
from battlemetrics.models.user import User

if TYPE_CHECKING:
    from asyncio import AbstractEventLoop
    from types import TracebackType

    from aiohttp import BaseConnector, BasicAuth

__all__ = ("Battlemetrics",)


class Battlemetrics:
    """The main client to handle all the Battlemetrics requests.

    Parameters
    ----------
        api_key (str)
            Your given API token.
    """

    def __init__(
        self,
        api_key: str,
        *,
        asyncio_debug: bool = False,
        connector: BaseConnector | None = None,
        loop: AbstractEventLoop | None = None,
        proxy: str | None = None,
        proxy_auth: BasicAuth | None = None,
    ) -> None:
        self.__api_key = api_key

        if loop is None:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", DeprecationWarning)
                self.loop: asyncio.AbstractEventLoop = asyncio.get_event_loop()
        else:
            self.loop: asyncio.AbstractEventLoop = loop

        self.loop.set_debug(asyncio_debug)

        self.http = HTTPClient(
            api_key=self.__api_key,
            connector=connector,
            loop=loop,
            proxy=proxy,
            proxy_auth=proxy_auth,
        )

    async def __aenter__(self) -> "Battlemetrics":
        """Enter the context manager and return the Battlemetrics client."""
        return self

    async def __aexit__(
        self,
        type: type[BaseException] | None,  # noqa: A002
        value: BaseException | None,
        tb: TracebackType | None,
    ) -> None:
        """Close the client when exiting the context."""
        await self.close()

    async def close(self) -> None:
        """Close the client."""
        await self.http.close()

    # Helpers / Getters

    async def create_ban(
        self,
        player_id: int,
        *,
        reason: str,
        note: str | None,
        banlist_id: str,
        organization_id: int,
        server_id: int,
        org_wide: bool = True,
        auto_add_enabled: bool = True,
        native_enabled: bool = True,
        identifiers: (
            list[str | dict[str, Any]] | None
        ) = None,  # TODO: Add player object
        expires: str | None = None,
    ) -> Ban:
        """Create a ban with all required and optional parameters."""
        resp = await self.http.create_ban(
            player_id=player_id,
            reason=reason,
            note=note,
            banlist_id=banlist_id,
            organization_id=organization_id,
            server_id=server_id,
            org_wide=org_wide,
            auto_add_enabled=auto_add_enabled,
            native_enabled=native_enabled,
            identifiers=identifiers,
            expires=expires,
        )
        return Ban.model_validate(resp["data"])

    async def import_bans(self, bans: list[dict[str, Any]]) -> list[dict[str, Any]]:
        """Import multiple bans at once.

        This method allows adding multiple bans at once with certain limitations.

        Parameters
        ----------
        bans : list[dict[str, Any]]
            A list of ban data dictionaries, each containing attributes and relationships.

        Returns
        -------
        Any
            The response from the API.

        Raises
        ------
        BMException
            Will raise if the request fails or the response indicates an error.
        """
        return await self.http.import_bans(bans)

    async def export_bans(
        self,
        file_format: Literal[
            "arma2/bans.txt",
            "arma3/bans.txt",
            "squad/Bans.cfg",
            "ark/banlist.txt",
            "rust/bans.cfg",
            "rust/bansip_SERVER.ini",
        ],
        *,
        organization_id: int | None = None,
        server_id: int | None = None,
    ) -> Any:
        """Export bans in a specific format.

        Parameters
        ----------
        file_format : str
            The format to export the bans in.

        Returns
        -------
        Any
            The response from the API containing the exported bans.

        Raises
        ------
        BMException
            Will raise if the request fails or the response indicates an error.
        """
        return await self.http.export_bans(
            file_format=file_format,
            organization_id=organization_id,
            server_id=server_id,
        )

    async def delete_ban(self, ban_id: int) -> None:
        """Delete a specific ban by its ID.

        Parameters
        ----------
        ban_id : int
            The ID of the ban to delete.


        Raises
        ------
        BMException
            Will raise if the request fails or the response indicates an error.
        """
        await self.http.delete_ban(ban_id)

    async def list_bans(
        self,
        *,
        banlist_id: str | None = None,
        exempt: bool | None = None,
        expired: bool | None = None,
        organization_id: int | None = None,
        player_id: int | None = None,
        search: str | None = None,
        server_id: int | None = None,
        user_ids: list[int] | None = None,
        page_size: int | None = None,
    ) -> list[Ban]:
        """List bans with optional filters for organization and server.

        Parameters
        ----------
        organization_id : int | None
            The ID of the organization to filter by.
        server_id : int | None
            The ID of the server to filter by.

        Returns
        -------
        Any
            The response from the API containing the list of bans.

        Raises
        ------
        BMException
            Will raise if the request fails or the response indicates an error.
        """
        resp = await self.http.list_bans(
            banlist_id=banlist_id,
            exempt=exempt,
            expired=expired,
            organization_id=organization_id,
            player_id=player_id,
            search=search,
            server_id=server_id,
            user_ids=user_ids,
            page_size=page_size,
        )
        return [Ban.model_validate(ban) for ban in resp["data"]]

    async def update_ban(
        self,
        ban_id: int | None = None,
        *,
        reason: str | None = None,
        note: str | None = None,
        identifiers: list[str | dict[str, Any]] | None = None,
        expires: str | None = None,
        org_wide: bool = True,
        auto_add_enabled: bool = True,
        native_enabled: bool = True,
    ) -> Ban:
        """Update a specific ban by its ID.

        Parameters
        ----------
        ban_id : int
            The ID of the ban to update.
        **kwargs : Any
            The fields to update.

        Raises
        ------
        BMException
            Will raise if the request fails or the response indicates an error.
        """
        resp = await self.http.update_ban(
            ban_id=ban_id,
            reason=reason,
            note=note,
            identifiers=identifiers,
            expires=expires,
            org_wide=org_wide,
            auto_add_enabled=auto_add_enabled,
            native_enabled=native_enabled,
        )
        return Ban.model_validate(resp["data"])

    async def ban_info(self, ban_id: int) -> Ban:
        """Get information about a specific ban."""
        resp = await self.http.ban_info(ban_id)
        return Ban.model_validate(resp["data"])

    async def create_banlist_exemption(
        self,
        ban_id: int,
        organization_id: int,
        *,
        reason: str | None = None,
    ) -> BanListExemption:
        """Create a banlist Exemption."""
        resp = await self.http.create_banlist_exemption(
            ban_id=ban_id,
            organization_id=organization_id,
            reason=reason,
        )
        return BanListExemption.model_validate(resp["data"])

    async def read_banlist_exemption(
        self,
        ban_id: int,
        ban_exemption_id: int,
    ) -> dict[str, Any]:
        """Read a specific banlist exemption by its ID.

        Parameters
        ----------
        ban_id : int
            The ID of the ban to read.
        ban_exemption_id : int
            The ID of the ban exemption to read.

        Returns
        -------
        dict[str, Any]
            The response from the API containing the banlist exemption.

        Raises
        ------
        BMException
            Will raise if the request fails or the response indicates an error.
        """
        resp = await self.http.read_banlist_exemption(
            ban_id=ban_id,
            ban_exemption_id=ban_exemption_id,
        )
        return BanListExemption.model_validate(resp["data"])

    async def list_banlist_exemptions(self, ban_id: int) -> dict[str, Any]:
        """List all banlist exemptions for a specific ban.

        Parameters
        ----------
        ban_id : int
            The ID of the ban to list exemptions for.

        Returns
        -------
        dict[str, Any]
            The response from the API containing the list of banlist exemptions.

        Raises
        ------
        BMException
            Will raise if the request fails or the response indicates an error.
        """
        resp = await self.http.list_banlist_exemptions(ban_id=ban_id)
        return [
            BanListExemption.model_validate(exemption) for exemption in resp["data"]
        ]

    async def update_banlist_exemption(
        self,
        ban_id: int,
        *,
        reason: str | None = None,
    ) -> dict[str, Any]:
        """Update a specific banlist exemption by its ID.

        Parameters
        ----------
        ban_id : int
            The ID of the ban to update.
        ban_exemption_id : int
            The ID of the ban exemption to update.
        reason : str | None
            The new reason for the ban exemption.

        Returns
        -------
        dict[str, Any]
            The response from the API containing the updated banlist exemption.

        Raises
        ------
        BMException
            Will raise if the request fails or the response indicates an error.
        """
        resp = await self.http.update_banlist_exemption(
            ban_id=ban_id,
            reason=reason,
        )
        return BanListExemption.model_validate(resp["data"])

    async def delete_banlist_exemption(
        self,
        ban_id: int,
    ) -> None:
        """Delete a specific banlist exemption by its ID.

        Parameters
        ----------
        ban_id : int
            The ID of the ban to delete the exemption from.
        ban_exemption_id : int
            The ID of the ban exemption to delete.

        Returns
        -------
        dict[str, Any]
            The response from the API confirming the deletion.

        Raises
        ------
        BMException
            Will raise if the request fails or the response indicates an error.
        """
        await self.http.delete_banlist_exemption(
            ban_id=ban_id,
        )

    # ----------------------------- Ban Lists ----------------------------- #

    async def create_banlist(
        self,
        *,
        name: str,
        action: Literal["none", "log", "kick"],
        organization_id: int,
        owner_id: int,
        default_identifiers: list[str] | None = None,
        default_reasons: list[str] | None = None,
        default_auto_add_enabled: bool = True,
        default_native_enabled: bool | None = None,
        native_ban_ttl: int | None = None,
        native_ban_temp_max_expires: int | None = None,
        native_ban_perm_max_expires: int | None = None,
    ) -> BanList:
        """Create a ban list."""
        resp = await self.http.create_banlist(
            name=name,
            action=action,
            organization_id=organization_id,
            owner_id=owner_id,
            default_identifiers=default_identifiers,
            default_reasons=default_reasons,
            default_auto_add_enabled=default_auto_add_enabled,
            default_native_enabled=default_native_enabled,
            native_ban_ttl=native_ban_ttl,
            native_ban_temp_max_expires=native_ban_temp_max_expires,
            native_ban_perm_max_expires=native_ban_perm_max_expires,
        )
        return BanList.model_validate(resp["data"])

    async def create_banlist_from_invite(
        self,
        *,
        code: str,
        action: Literal["none", "log", "kick"],
        organization_id: int,
        owner_id: int,
        default_identifiers: list[str] | None = None,
        default_reasons: list[str] | None = None,
        default_auto_add_enabled: bool = True,
        default_native_enabled: bool | None = None,
        native_ban_ttl: int | None = None,
        native_ban_temp_max_expires: int | None = None,
        native_ban_perm_max_expires: int | None = None,
    ) -> BanList:
        """Accept a ban list invite and subscribe."""
        resp = await self.http.create_banlist_from_invite(
            code=code,
            action=action,
            organization_id=organization_id,
            owner_id=owner_id,
            default_identifiers=default_identifiers,
            default_reasons=default_reasons,
            default_auto_add_enabled=default_auto_add_enabled,
            default_native_enabled=default_native_enabled,
            native_ban_ttl=native_ban_ttl,
            native_ban_temp_max_expires=native_ban_temp_max_expires,
            native_ban_perm_max_expires=native_ban_perm_max_expires,
        )
        return BanList.model_validate(resp["data"])

    async def list_banlists(
        self,
        *,
        include: str | None = None,
        page_size: int | None = None,
    ) -> list[BanList]:
        """List ban lists owned or subscribed to."""
        resp = await self.http.list_banlists(include=include, page_size=page_size)
        return [BanList.model_validate(b) for b in resp["data"]]

    async def get_banlist(
        self,
        banlist_id: str,
        *,
        include: str | None = None,
    ) -> BanList:
        """Get a ban list."""
        resp = await self.http.get_banlist(banlist_id, include=include)
        return BanList.model_validate(resp["data"])

    async def update_banlist(
        self,
        banlist_id: str,
        *,
        name: str | None = None,
        action: Literal["none", "log", "kick"] | None = None,
        default_identifiers: list[str] | None = None,
        default_reasons: list[str] | None = None,
        default_auto_add_enabled: bool | None = None,
        default_native_enabled: bool | None = None,
        native_ban_ttl: int | None = None,
        native_ban_temp_max_expires: int | None = None,
        native_ban_perm_max_expires: int | None = None,
    ) -> BanList:
        """Update a ban list."""
        resp = await self.http.update_banlist(
            banlist_id=banlist_id,
            name=name,
            action=action,
            default_identifiers=default_identifiers,
            default_reasons=default_reasons,
            default_auto_add_enabled=default_auto_add_enabled,
            default_native_enabled=default_native_enabled,
            native_ban_ttl=native_ban_ttl,
            native_ban_temp_max_expires=native_ban_temp_max_expires,
            native_ban_perm_max_expires=native_ban_perm_max_expires,
        )
        return BanList.model_validate(resp["data"])

    async def list_banlist_organizations(
        self,
        banlist_id: str,
        *,
        include: str | None = None,
        page_size: int | None = None,
    ) -> list[Organization]:
        """List organizations subscribed to a ban list."""
        resp = await self.http.list_banlist_organizations(
            banlist_id=banlist_id,
            include=include,
            page_size=page_size,
        )
        return [Organization.model_validate(o) for o in resp["data"]]

    async def get_banlist_subscription(
        self,
        banlist_id: str,
        organization_id: int,
        *,
        include: str | None = None,
    ) -> BanList:
        """Get an organization's ban list subscription."""
        resp = await self.http.get_banlist_subscription(
            banlist_id=banlist_id,
            organization_id=organization_id,
            include=include,
        )
        return BanList.model_validate(resp["data"])

    async def update_banlist_subscription(
        self,
        banlist_id: str,
        organization_id: int,
        *,
        perm_manage: bool | None = None,
        perm_create: bool | None = None,
        perm_update: bool | None = None,
        perm_delete: bool | None = None,
    ) -> BanList:
        """Update an organization's ban list subscription permissions."""
        resp = await self.http.update_banlist_subscription(
            banlist_id=banlist_id,
            organization_id=organization_id,
            perm_manage=perm_manage,
            perm_create=perm_create,
            perm_update=perm_update,
            perm_delete=perm_delete,
        )
        return BanList.model_validate(resp["data"])

    async def unsubscribe_banlist(self, banlist_id: str, organization_id: int) -> None:
        """Unsubscribe an organization from a ban list."""
        await self.http.unsubscribe_banlist(banlist_id, organization_id)

    # -------------------------- Ban List Invites ------------------------- #

    async def create_banlist_invite(
        self,
        banlist_id: str,
        *,
        limit: int | None = None,
        perm_manage: bool | None = None,
        perm_create: bool | None = None,
        perm_update: bool | None = None,
        perm_delete: bool | None = None,
        uses: int | None = None,
        organization_id: int | None = None,
    ) -> BanListInvite:
        """Create a ban list invite."""
        resp = await self.http.create_banlist_invite(
            banlist_id=banlist_id,
            limit=limit,
            perm_manage=perm_manage,
            perm_create=perm_create,
            perm_update=perm_update,
            perm_delete=perm_delete,
            uses=uses,
            organization_id=organization_id,
        )
        return BanListInvite.model_validate(resp["data"])

    async def list_banlist_invites(
        self,
        banlist_id: str,
        *,
        include: str | None = None,
    ) -> list[BanListInvite]:
        """List invites for a ban list."""
        resp = await self.http.list_banlist_invites(
            banlist_id=banlist_id,
            include=include,
        )
        return [BanListInvite.model_validate(i) for i in resp["data"]]

    async def get_banlist_invite(
        self,
        invite_id: str,
        *,
        include: str | None = None,
    ) -> BanListInvite:
        """Get a ban list invite."""
        resp = await self.http.get_banlist_invite(invite_id=invite_id, include=include)
        return BanListInvite.model_validate(resp["data"])

    async def delete_banlist_invite(self, invite_id: str) -> None:
        """Delete a ban list invite."""
        await self.http.delete_banlist_invite(invite_id)

    # ----------------------------- Native Bans ---------------------------- #

    async def list_native_bans(
        self,
        *,
        ban_id: int | None = None,
        server_id: int | None = None,
        include: str | None = None,
        page_size: int | None = None,
        sort: str | None = None,
    ) -> list[NativeBan]:
        """List native (game) bans."""
        resp = await self.http.list_native_bans(
            ban_id=ban_id,
            server_id=server_id,
            include=include,
            page_size=page_size,
            sort=sort,
        )
        return [NativeBan.model_validate(nb) for nb in resp["data"]]

    # --------------------------- Commands Activity ------------------------ #

    async def list_command_stats(
        self,
        organization_id: int,
        *,
        commands: list[str] | None = None,
        servers: list[int] | None = None,
        users: list[int] | None = None,
        timestamp_range: str | None = None,
        summary: bool | None = None,
    ) -> list[CommandsActivity]:
        """List command activity statistics."""
        resp = await self.http.list_command_stats(
            organization_id=organization_id,
            commands=commands,
            servers=servers,
            users=users,
            timestamp_range=timestamp_range,
            summary=summary,
        )
        return [CommandsActivity.model_validate(c) for c in resp["data"]]

    # -------------------------------- Coplay ------------------------------ #

    async def list_coplay(
        self,
        player_id: int,
        *,
        period: str,
        players: list[int] | None = None,
        servers: list[int] | None = None,
        organizations: list[int] | None = None,
        page_size: int | None = None,
    ) -> list[Coplay]:
        """List coplay entries for a player."""
        resp = await self.http.list_coplay(
            player_id=player_id,
            period=period,
            players=players,
            servers=servers,
            organizations=organizations,
            page_size=page_size,
        )
        return [Coplay.model_validate(cp) for cp in resp["data"]]

    # ------------------------------ Player Flags -------------------------- #

    async def create_player_flag(
        self,
        *,
        organization_id: int,
        name: str,
        color: str | None = None,
        icon: str | None = None,
        description: str | None = None,
    ) -> PlayerFlag:
        """Create a player flag."""
        resp = await self.http.create_player_flag(
            organization_id=organization_id,
            name=name,
            color=color,
            icon=icon,
            description=description,
        )
        return PlayerFlag.model_validate(resp["data"])

    async def list_player_flags(
        self,
        *,
        personal: bool | None = None,
        include: str | None = None,
        page_size: int | None = None,
    ) -> list[PlayerFlag]:
        """List player flags."""
        resp = await self.http.list_player_flags(
            personal=personal,
            include=include,
            page_size=page_size,
        )
        return [PlayerFlag.model_validate(f) for f in resp["data"]]

    async def update_player_flag(
        self,
        flag_id: str,
        *,
        name: str | None = None,
        color: str | None = None,
        icon: str | None = None,
        description: str | None = None,
    ) -> PlayerFlag:
        """Update a player flag."""
        resp = await self.http.update_player_flag(
            flag_id=flag_id,
            name=name,
            color=color,
            icon=icon,
            description=description,
        )
        return PlayerFlag.model_validate(resp["data"])

    async def delete_player_flag(self, flag_id: str) -> None:
        """Delete a player flag."""
        await self.http.delete_player_flag(flag_id)

    async def list_player_flag_assignments(
        self,
        player_id: int,
        *,
        include: str | None = None,
        page_size: int | None = None,
    ) -> list[FlagPlayer]:
        """List player flag assignments for a player."""
        resp = await self.http.list_player_flag_assignments(
            player_id=player_id,
            include=include,
            page_size=page_size,
        )
        return [FlagPlayer.model_validate(a) for a in resp["data"]]

    async def remove_player_flag_assignment(
        self,
        player_id: int,
        player_flag_id: str,
    ) -> None:
        """Remove a player flag assignment."""
        await self.http.remove_player_flag_assignment(player_id, player_flag_id)

    # --------------------------------- Games ------------------------------ #

    async def list_games(self, *, page_size: int | None = None) -> list[Game]:
        """List games."""
        resp = await self.http.list_games(page_size=page_size)
        return [Game.model_validate(g) for g in resp["data"]]

    async def get_game(self, game_id: str) -> Game:
        """Get a game."""
        resp = await self.http.get_game(game_id)
        return Game.model_validate(resp["data"])

    # ---------------------------- Game Features --------------------------- #

    async def list_game_features(
        self,
        *,
        game: str | None = None,
        page_size: int | None = None,
    ) -> list[GameFeatures]:
        """List game features."""
        resp = await self.http.list_game_features(game=game, page_size=page_size)
        return [GameFeatures.model_validate(f) for f in resp["data"]]

    async def list_game_feature_options(
        self,
        feature_id: str,
        *,
        page_size: int | None = None,
        sort: str | None = None,
    ) -> list[GameFeatureOptions]:
        """List game feature options."""
        resp = await self.http.list_game_feature_options(
            feature_id=feature_id,
            page_size=page_size,
            sort=sort,
        )
        return [GameFeatureOptions.model_validate(o) for o in resp["data"]]

    # ------------------------------ Leaderboards ------------------------- #

    async def get_leaderboard(
        self,
        server_id: int,
        *,
        period: str,
        player_id: int | None = None,
        leaderboard: str = "time",
        page_size: int | None = None,
        offset: int | None = None,
    ) -> list[Leaderboard]:
        """Get a server leaderboard."""
        resp = await self.http.get_leaderboard(
            server_id=server_id,
            period=period,
            player_id=player_id,
            leaderboard=leaderboard,
            page_size=page_size,
            offset=offset,
        )
        return [Leaderboard.model_validate(lb) for lb in resp["data"]]

    # ---------------------------- Organization Stats --------------------- #

    async def get_organization_player_stats(
        self,
        organization_id: int,
        *,
        range_: str,
        game: str | None = None,
        group: str | None = None,
    ) -> OrganizationStats:
        """Get organization player stats."""
        resp = await self.http.get_organization_player_stats(
            organization_id=organization_id,
            range_=range_,
            game=game,
            group=group,
        )
        return OrganizationStats.model_validate(resp["data"])

    # -------------------------------- Players ----------------------------- #

    async def list_players(
        self,
        *,
        search: str | None = None,
        servers: list[int] | None = None,
        organizations: list[int] | None = None,
        game: str | None = None,
        page_size: int | None = None,
        sort: str | None = None,
        online: bool | None = None,
    ) -> list[Player]:
        """List players."""
        resp = await self.http.list_players(
            search=search,
            servers=servers,
            organizations=organizations,
            game=game,
            page_size=page_size,
            sort=sort,
            online=online,
        )
        return [Player.model_validate(p) for p in resp["data"]]

    async def get_player(self, player_id: int, *, include: str | None = None) -> Player:
        """Get a player."""
        resp = await self.http.get_player(player_id, include=include)
        return Player.model_validate(resp["data"])

    async def match_players(self, identifiers: list[dict[str, str]]) -> list[Player]:
        """Match players (slow)."""
        resp = await self.http.match_players(identifiers)
        return [Player.model_validate(p) for p in resp["data"]]

    async def quick_match_players(
        self,
        identifiers: list[dict[str, str]],
    ) -> list[QuickMatchIdentifier]:
        """Quick match players by identifiers.

        This API method is rate limited to 10 requests per second.
        Enterprise users have a higher rate limit of 30 requests per second.

        Parameters
        ----------
        identifiers : list[dict[str, str]]
            A list of identifier dicts with 'type' and 'identifier' keys.
            Example: [{"type": "steamID", "identifier": "76561197960265720"}]

        Returns
        -------
        list[QuickMatchIdentifier]
            A list of matched identifiers with player relationships.
        """
        resp = await self.http.quick_match_players(identifiers)
        return [QuickMatchIdentifier.model_validate(p) for p in resp["data"]]

    async def player_session_history(
        self,
        player_id: int,
        *,
        servers: list[int] | None = None,
        organizations: list[int] | None = None,
        page_size: int | None = None,
        include: str | None = None,
    ) -> list[Session]:
        """Get a player's session history."""
        resp = await self.http.player_session_history(
            player_id=player_id,
            servers=servers,
            organizations=organizations,
            page_size=page_size,
            include=include,
        )
        return [Session.model_validate(s) for s in resp["data"]]

    async def related_identifiers(
        self,
        player_id: int,
        *,
        match_identifiers: list[str] | None = None,
        identifiers: list[str] | None = None,
        page_size: int | None = None,
        include: str | None = None,
    ) -> list[RelatedPlayerIdentifier]:
        """Get related identifiers."""
        resp = await self.http.related_identifiers(
            player_id=player_id,
            match_identifiers=match_identifiers,
            identifiers=identifiers,
            page_size=page_size,
            include=include,
        )
        return [RelatedPlayerIdentifier.model_validate(r) for r in resp["data"]]

    # ----------------------------- Player Queries ------------------------- #

    async def list_player_queries(
        self,
        *,
        organizations: list[int] | None = None,
        ids: list[str] | None = None,
        sort: str | None = None,
        page_size: int | None = None,
    ) -> list[PlayerQuery]:
        """List player queries."""
        resp = await self.http.list_player_queries(
            organizations=organizations,
            ids=ids,
            sort=sort,
            page_size=page_size,
        )
        return [PlayerQuery.model_validate(q) for q in resp["data"]]

    async def get_player_query(self, query_id: str) -> PlayerQuery:
        """Get a player query."""
        resp = await self.http.get_player_query(query_id)
        return PlayerQuery.model_validate(resp["data"])

    async def create_player_query(
        self,
        organization_id: int,
        *,
        query_name: str,
        conditions: list[dict[str, Any]],
    ) -> PlayerQuery:
        """Create a player query."""
        resp = await self.http.create_player_query(
            organization_id=organization_id,
            query_name=query_name,
            conditions=conditions,
        )
        return PlayerQuery.model_validate(resp["data"])

    async def update_player_query(
        self,
        query_id: str,
        *,
        query_name: str | None = None,
        conditions: list[dict[str, Any]] | None = None,
    ) -> PlayerQuery:
        """Update a player query."""
        resp = await self.http.update_player_query(
            query_id=query_id,
            query_name=query_name,
            conditions=conditions,
        )
        return PlayerQuery.model_validate(resp["data"])

    async def delete_player_query(self, query_id: str) -> None:
        """Delete a player query."""
        await self.http.delete_player_query(query_id)

    async def run_player_query(
        self,
        player_id: int,
        query_id: str,
        *,
        identifiers: list[str] | None = None,
        include: str | None = None,
        page_size: int | None = None,
        offset: int | None = None,
    ) -> list[PlayerQueryResult]:
        """Run a saved player query for a player."""
        resp = await self.http.run_player_query(
            player_id=player_id,
            query_id=query_id,
            identifiers=identifiers,
            include=include,
            page_size=page_size,
            offset=offset,
        )
        return [PlayerQueryResult.model_validate(r) for r in resp["data"]]

    # ----------------------------- Reserved Slots ------------------------- #

    async def create_reserved_slot(
        self,
        *,
        player_id: int,
        organization_id: int,
        server_ids: list[int],
        identifiers: list[str | dict[str, Any]] | None = None,
        expires: str | None = None,
    ) -> ReservedSlot:
        """Create a reserved slot."""
        resp = await self.http.create_reserved_slot(
            player_id=player_id,
            organization_id=organization_id,
            server_ids=server_ids,
            identifiers=identifiers,
            expires=expires,
        )
        return ReservedSlot.model_validate(resp["data"])

    async def get_reserved_slot(
        self,
        slot_id: str,
        *,
        include: str | None = None,
    ) -> ReservedSlot:
        """Get a reserved slot."""
        resp = await self.http.get_reserved_slot(slot_id, include=include)
        return ReservedSlot.model_validate(resp["data"])

    async def list_reserved_slots(
        self,
        *,
        organization_id: int | None = None,
        player_id: int | None = None,
        server_id: int | None = None,
        search: str | None = None,
        expired: bool | None = None,
        include: str | None = None,
        page_size: int | None = None,
    ) -> list[ReservedSlot]:
        """List reserved slots."""
        resp = await self.http.list_reserved_slots(
            organization_id=organization_id,
            player_id=player_id,
            server_id=server_id,
            search=search,
            expired=expired,
            include=include,
            page_size=page_size,
        )
        return [ReservedSlot.model_validate(rs) for rs in resp["data"]]

    async def update_reserved_slot(
        self,
        slot_id: str,
        *,
        identifiers: list[str | dict[str, Any]] | None = None,
        expires: str | None = None,
        server_ids: list[int] | None = None,
    ) -> ReservedSlot:
        """Update a reserved slot."""
        resp = await self.http.update_reserved_slot(
            slot_id=slot_id,
            identifiers=identifiers,
            expires=expires,
            server_ids=server_ids,
        )
        return ReservedSlot.model_validate(resp["data"])

    async def delete_reserved_slot(self, slot_id: str) -> None:
        """Delete a reserved slot."""
        await self.http.delete_reserved_slot(slot_id)

    # -------------------------------- Servers ----------------------------- #

    async def list_servers(
        self,
        *,
        search: str | None = None,
        game: str | None = None,
        status: str | None = None,
        countries: list[str] | None = None,
        page_size: int | None = None,
        sort: str | None = None,
    ) -> list[Server]:
        """List servers."""
        resp = await self.http.list_servers(
            search=search,
            game=game,
            status=status,
            countries=countries,
            page_size=page_size,
            sort=sort,
        )
        return [Server.model_validate(s) for s in resp["data"]]

    async def get_server(self, server_id: int, *, include: str | None = None) -> Server:
        """Get a server."""
        resp = await self.http.get_server(server_id, include=include)
        return Server.model_validate(resp["data"])

    async def server_sessions(
        self,
        server_id: int,
        *,
        start: str | None = None,
        stop: str | None = None,
        at: str | None = None,
        include: str | None = None,
    ) -> list[Session]:
        """List sessions for a server."""
        resp = await self.http.server_sessions(
            server_id=server_id,
            start=start,
            stop=stop,
            at=at,
            include=include,
        )
        return [Session.model_validate(s) for s in resp["data"]]

    async def server_outages(
        self,
        server_id: int,
        *,
        range_: str | None = None,
        include: str | None = None,
        page_size: int | None = None,
    ) -> Any:
        """Return raw outage data (no model)."""
        return await self.http.server_outages(
            server_id=server_id,
            range_=range_,
            include=include,
            page_size=page_size,
        )

    # -------------------------------- Sessions ---------------------------- #

    async def list_sessions(
        self,
        *,
        servers: list[int] | None = None,
        games: list[str] | None = None,
        organizations: list[int] | None = None,
        players: list[int] | None = None,
        identifiers: list[str] | None = None,
        range_: str | None = None,
        at: str | None = None,
        include: str | None = None,
        page_size: int | None = None,
    ) -> list[Session]:
        """List sessions."""
        resp = await self.http.list_sessions(
            servers=servers,
            games=games,
            organizations=organizations,
            players=players,
            identifiers=identifiers,
            range_=range_,
            at=at,
            include=include,
            page_size=page_size,
        )
        return [Session.model_validate(s) for s in resp["data"]]

    async def session_coplay(
        self,
        session_id: str,
        *,
        page_size: int | None = None,
        include: str | None = None,
    ) -> list[Coplay]:
        """Get coplay data for a session."""
        resp = await self.http.session_coplay(
            session_id=session_id,
            page_size=page_size,
            include=include,
        )
        return [Coplay.model_validate(c) for c in resp["data"]]

    # --------------------------------- Users ------------------------------ #

    async def get_user(self, user_id: int) -> User:
        """Get a user."""
        resp = await self.http.get_user(user_id)
        return User.model_validate(resp["data"])

    # -------------------------------- Player Notes ----------------------- #

    async def list_player_notes(
        self,
        player_id: int,
        *,
        expired: bool | None = None,
        organizations: list[int] | None = None,
        personal: bool | None = None,
        search: str | None = None,
        users: list[int] | None = None,
        include: str | None = None,
        page_size: int | None = None,
        page_key: str | None = None,
        page_rel: Literal["next", "prev"] | None = None,
    ) -> list[Note]:
        """List player notes."""
        resp = await self.http.list_player_notes(
            player_id=player_id,
            expired=expired,
            organizations=organizations,
            personal=personal,
            search=search,
            users=users,
            include=include,
            page_size=page_size,
            page_key=page_key,
            page_rel=page_rel,
        )
        return [Note.model_validate(n) for n in resp["data"]]

    async def get_player_note(self, player_id: int, note_id: str) -> Note:
        """Get a player note."""
        resp = await self.http.get_player_note(player_id, note_id)
        return Note.model_validate(resp["data"])

    async def create_player_note(
        self,
        player_id: int,
        *,
        note: str,
        shared: bool,
        expires_at: str | None = None,
        clearance_level: int | None = None,
        organization_id: int | None = None,
        trigger_id: str | None = None,
    ) -> Note:
        """Create a player note."""
        resp = await self.http.create_player_note(
            player_id=player_id,
            note=note,
            shared=shared,
            expires_at=expires_at,
            clearance_level=clearance_level,
            organization_id=organization_id,
            trigger_id=trigger_id,
        )
        return Note.model_validate(resp["data"])

    async def update_player_note(
        self,
        player_id: int,
        note_id: str,
        *,
        note: str | None = None,
        shared: bool | None = None,
        expires_at: str | None = None,
        clearance_level: int | None = None,
    ) -> Note:
        """Update a player note."""
        resp = await self.http.update_player_note(
            player_id=player_id,
            note_id=note_id,
            note=note,
            shared=shared,
            expires_at=expires_at,
            clearance_level=clearance_level,
        )
        return Note.model_validate(resp["data"])

    async def delete_player_note(self, player_id: int, note_id: str) -> None:
        """Delete a player note."""
        await self.http.delete_player_note(player_id, note_id)

    # -------------------------- Organization Friends ----------------------- #

    async def list_organization_friends(
        self,
        organization_id: int,
        *,
        include: str | None = None,
        page_size: int | None = None,
    ) -> list[OrganizationFriend]:
        """List organization friends."""
        resp = await self.http.list_organization_friends(
            organization_id=organization_id,
            include=include,
            page_size=page_size,
        )
        return [OrganizationFriend.model_validate(f) for f in resp["data"]]

    async def get_organization_friend(
        self,
        organization_id: int,
        friend_id: int,
        *,
        include: str | None = None,
    ) -> OrganizationFriend:
        """Get a specific organization friend."""
        resp = await self.http.get_organization_friend(
            organization_id=organization_id,
            friend_id=friend_id,
            include=include,
        )
        return OrganizationFriend.model_validate(resp["data"])

    async def create_organization_friend(
        self,
        organization_id: int,
        friend_id: int,
        *,
        identifiers: list[str] | None = None,
        notes: bool = False,
        player_flags: list[str] | None = None,
    ) -> OrganizationFriend:
        """Create an organization friend request."""
        resp = await self.http.create_organization_friend(
            organization_id=organization_id,
            friend_id=friend_id,
            identifiers=identifiers,
            notes=notes,
            player_flags=player_flags,
        )
        return OrganizationFriend.model_validate(resp["data"])

    async def update_organization_friend(
        self,
        organization_id: int,
        friend_id: int,
        *,
        identifiers: list[str] | None = None,
        notes: bool | None = None,
        player_flags: list[str] | None = None,
    ) -> OrganizationFriend:
        """Update an organization friend."""
        resp = await self.http.update_organization_friend(
            organization_id=organization_id,
            friend_id=friend_id,
            identifiers=identifiers,
            notes=notes,
            player_flags=player_flags,
        )
        return OrganizationFriend.model_validate(resp["data"])

    async def delete_organization_friend(
        self,
        organization_id: int,
        friend_id: int,
    ) -> None:
        """Delete an organization friend."""
        await self.http.delete_organization_friend(organization_id, friend_id)

    async def bulk_delete_organization_friends(
        self,
        organization_id: int,
        friend_ids: list[int],
    ) -> None:
        """Bulk delete organization friends."""
        await self.http.bulk_delete_organization_friends(organization_id, friend_ids)

    # ------------------------ Additional Server Methods --------------------- #

    async def server_downtime(
        self,
        server_id: int,
        *,
        start: str,
        stop: str,
        include: str | None = None,
        page_size: int | None = None,
    ) -> Any:
        """Get server downtime history."""
        return await self.http.server_downtime(
            server_id=server_id,
            start=start,
            stop=stop,
            include=include,
            page_size=page_size,
        )

    async def server_first_time_history(
        self,
        server_id: int,
        *,
        start: str,
        stop: str,
    ) -> Any:
        """Get server first-time player history."""
        return await self.http.server_first_time_history(
            server_id=server_id,
            start=start,
            stop=stop,
        )

    async def server_player_count_history(
        self,
        server_id: int,
        *,
        start: str,
        stop: str,
        resolution: Literal["raw", "30", "60", "1440"] | None = None,
    ) -> Any:
        """Get server player count history."""
        return await self.http.server_player_count_history(
            server_id=server_id,
            start=start,
            stop=stop,
            resolution=resolution,
        )

    async def server_rank_history(
        self,
        server_id: int,
        *,
        start: str,
        stop: str,
    ) -> Any:
        """Get server rank history."""
        return await self.http.server_rank_history(
            server_id=server_id,
            start=start,
            stop=stop,
        )

    async def server_group_rank_history(
        self,
        server_id: int,
        *,
        start: str,
        stop: str,
    ) -> Any:
        """Get server group rank history."""
        return await self.http.server_group_rank_history(
            server_id=server_id,
            start=start,
            stop=stop,
        )

    async def server_time_played_history(
        self,
        server_id: int,
        *,
        start: str,
        stop: str,
    ) -> Any:
        """Get server time played history."""
        return await self.http.server_time_played_history(
            server_id=server_id,
            start=start,
            stop=stop,
        )

    async def server_unique_player_history(
        self,
        server_id: int,
        *,
        start: str,
        stop: str,
    ) -> Any:
        """Get server unique player history."""
        return await self.http.server_unique_player_history(
            server_id=server_id,
            start=start,
            stop=stop,
        )

    async def create_server(
        self,
        *,
        game: str,
        ip: str,
        port: int,
        port_query: int,
        organization_id: int,
    ) -> Server:
        """Create a server."""
        resp = await self.http.create_server(
            game=game,
            ip=ip,
            port=port,
            port_query=port_query,
            organization_id=organization_id,
        )
        return Server.model_validate(resp["data"])

    async def update_server(
        self,
        server_id: int,
        *,
        metadata: dict[str, Any] | None = None,
        port_query: int | None = None,
        default_ban_list_id: str | None = None,
        server_group_id: str | None = None,
    ) -> Server:
        """Update a server."""
        resp = await self.http.update_server(
            server_id=server_id,
            metadata=metadata,
            port_query=port_query,
            default_ban_list_id=default_ban_list_id,
            server_group_id=server_group_id,
        )
        return Server.model_validate(resp["data"])

    async def enable_server_rcon(
        self,
        server_id: int,
        *,
        password: str,
        port: int | None = None,
        ip: str | None = None,
    ) -> Server:
        """Enable RCON for a server."""
        resp = await self.http.enable_server_rcon(
            server_id=server_id,
            password=password,
            port=port,
            ip=ip,
        )
        return Server.model_validate(resp["data"])

    async def delete_server_rcon(self, server_id: int) -> None:
        """Delete RCON configuration for a server."""
        await self.http.delete_server_rcon(server_id)

    async def disconnect_server_rcon(self, server_id: int) -> None:
        """Disconnect RCON for a server."""
        await self.http.disconnect_server_rcon(server_id)

    async def connect_server_rcon(self, server_id: int) -> None:
        """Connect RCON for a server."""
        await self.http.connect_server_rcon(server_id)

    async def force_update_server(self, server_id: int) -> Server:
        """Force update a server."""
        resp = await self.http.force_update_server(server_id)
        return Server.model_validate(resp["data"])

    # ------------------------- Additional Player Methods -------------------- #

    async def player_time_played_history(
        self,
        player_id: int,
        server_id: int,
        *,
        start: str,
        stop: str,
    ) -> Any:
        """Get a player's time played history for a server."""
        return await self.http.player_time_played_history(
            player_id=player_id,
            server_id=server_id,
            start=start,
            stop=stop,
        )

    async def get_player_server_info(
        self,
        player_id: int,
        server_id: int,
    ) -> Player:
        """Get player information for a specific server."""
        resp = await self.http.get_player_server_info(
            player_id=player_id,
            server_id=server_id,
        )
        return Player.model_validate(resp["data"])

    # ------------------------- Additional Flag Methods ---------------------- #

    async def get_player_flag(
        self,
        flag_id: str,
        *,
        include: str | None = None,
    ) -> PlayerFlag:
        """Get a specific player flag."""
        resp = await self.http.get_player_flag(flag_id, include=include)
        return PlayerFlag.model_validate(resp["data"])

    async def add_player_flag_assignment(
        self,
        player_id: int,
        player_flag_id: str,
    ) -> FlagPlayer:
        """Add a flag to a player."""
        resp = await self.http.add_player_flag_assignment(player_id, player_flag_id)
        return FlagPlayer.model_validate(resp["data"])

    # ----------------------- Additional Native Ban Methods ------------------ #

    async def force_update_native_ban(self, native_ban_id: str) -> NativeBan:
        """Force update a native ban."""
        resp = await self.http.force_update_native_ban(native_ban_id)
        return NativeBan.model_validate(resp["data"])

    # ------------------------ Additional Session Methods -------------------- #

    async def get_session(
        self,
        session_id: str,
        *,
        include: str | None = None,
    ) -> Session:
        """Get a specific session."""
        resp = await self.http.get_session(session_id, include=include)
        return Session.model_validate(resp["data"])

    # --------------------- Additional Player Query Methods ------------------ #

    async def run_custom_player_query(
        self,
        player_id: int,
        *,
        conditions: list[dict[str, Any]],
        identifiers: list[str] | None = None,
        include: str | None = None,
        page_size: int | None = None,
        offset: int | None = None,
    ) -> list[PlayerQueryResult]:
        """Run a custom related player query (not saved) for a player."""
        resp = await self.http.run_custom_player_query(
            player_id=player_id,
            conditions=conditions,
            identifiers=identifiers,
            include=include,
            page_size=page_size,
            offset=offset,
        )
        return [PlayerQueryResult.model_validate(r) for r in resp["data"]]



from typing import Literal

from pydantic import BaseModel, Field

from .base import Base, BaseRelationships, Relationship

PlayerFlagIcon = Literal[
    None,
    "flag",
    "alarm",
    "attach_money",
    "block",
    "bookmark",
    "check_circle_outline",
    "emoji_objects",
    "enhanced_encryption",
    "error_outline",
    "highlight_off",
    "info",
    "label",
    "loyalty",
    "monetization_on",
    "note_add",
    "notifications",
    "notification_important",
    "policy",
    "verified_user",
    "priority_high",
    "remove_circle",
    "report",
    "star",
    "thumb_down",
    "thumb_up",
    "visibility",
    "warning",
    "whatshot",
]


class FlagPlayerAttributes(BaseModel):
    """Attributes for the FlagPlayer model."""

    added_at: str = Field(alias="addedAt")
    removed_at: str | None = Field(default=None, alias="removedAt")

    model_config = {
        "populate_by_name": True,
    }


class FlagPlayerRelationships(BaseRelationships):
    """Relationships for the FlagPlayer model."""

    organization: Relationship | None = None
    player: Relationship | None = None
    player_flag: Relationship | None = Field(default=None, alias="playerFlag")
    user: Relationship | None = None


class FlagPlayer(Base):
    """Represents the relationship between a player flag and a player."""

    type: str = "flagPlayer"
    attributes: FlagPlayerAttributes
    relationships: FlagPlayerRelationships


class PlayerFlagAttributes(BaseModel):
    """Attributes for the PlayerFlag model."""

    color: str
    created_at: str = Field(alias="createdAt")
    description: str | None = None
    icon: PlayerFlagIcon = None
    name: str
    updated_at: str = Field(alias="updatedAt")

    model_config = {
        "populate_by_name": True,
    }


class PlayerFlagMeta(BaseModel):
    """Metadata for the PlayerFlag model."""

    shared: bool


class PlayerFlagRelationships(BaseRelationships):
    """Relationships for the PlayerFlag model."""

    organization: Relationship | None = None
    user: Relationship | None = None


class PlayerFlag(Base):
    """Player flag model representing a flag that can be assigned to players."""

    type: str = "playerFlag"
    attributes: PlayerFlagAttributes
    meta: PlayerFlagMeta | None = None
    relationships: PlayerFlagRelationships



from typing import Any, Literal

from pydantic import BaseModel, Field

from .base import Base, BaseRelationships, IdentifierTypesLiteral, Relationship


class BanAttributes(BaseModel):
    """Attributes for the Ban model."""

    reason: str
    note: str | None = None
    expires: str | None = None
    identifiers: list[str | dict[str, Any]]
    org_wide: bool = Field(alias="orgWide")
    auto_add_enabled: bool = Field(alias="autoAddEnabled")
    native_enabled: bool | None = Field(default=None, alias="nativeEnabled")
    timestamp: str
    uid: str
    id: int


class BanRelationships(BaseRelationships):
    """Relationships for the Ban model."""

    organization: Relationship
    server: Relationship
    player: Relationship
    ban_list: Relationship = Field(alias="banList")
    user: Relationship | None = None


class Ban(Base):
    """The Ban model representing a ban in Battlemetrics."""

    type: str = "ban"
    attributes: BanAttributes
    relationships: BanRelationships


class NativeBanAttributes(BaseModel):
    """Attributes for the NativeBan model."""

    created_at: str = Field(alias="createdAt")
    expires: str | None = None
    identifier: str
    reason: str | None = None
    state: Literal["added", "removed"]
    type: IdentifierTypesLiteral  # type: ignore[reportInvalidTypeForm]
    updated_at: str = Field(alias="updatedAt")

    model_config = {
        "populate_by_name": True,
    }


class NativeBanRelationships(BaseRelationships):
    """Relationships for the NativeBan model."""

    ban: Relationship
    organization: Relationship
    server: Relationship


class NativeBan(Base):
    """Represents a native ban in Battlemetrics."""

    type: str = "banNative"
    attributes: NativeBanAttributes
    relationships: NativeBanRelationships



from pydantic import BaseModel, Field

from .base import Base, BaseRelationships, IdentifierTypesLiteral, Relationship
from .server import ServerData


class PlayerAttributes(BaseModel):
    """Attributes for the Player model."""

    created_at: str = Field(alias="createdAt")
    id: str
    name: str
    positive_match: bool = Field(alias="positiveMatch")
    private: bool
    updated_at: str = Field(alias="updatedAt")

    model_config = {
        "populate_by_name": True,
    }


class PlayerMeta(BaseModel):
    """Metadata for the Player model."""

    class Metadata(BaseModel):
        """Metadata for the Player model."""

        key: str
        private: bool
        value: str | None = None

    metadata: list[Metadata] | None = None


class PlayerRelationships(BaseRelationships):
    """Relationships for the Player model."""

    organizations: list[Relationship] | None = None
    server: Relationship | None = None
    servers: list[ServerData] | None = None
    user: Relationship | None = None


class Player(Base):
    """Player model representing a player in Battlemetrics."""

    type: str = "player"
    attributes: PlayerAttributes
    meta: PlayerMeta | None = None
    relationships: PlayerRelationships


class PlayerIdentifierAttributes(BaseModel):
    """Attributes for the PlayerIdentifier model."""

    identifier: str
    last_seen: str = Field(alias="lastSeen")
    metadata: object | None = None
    private: bool
    type: IdentifierTypesLiteral  # type: ignore[reportInvalidTypeForm]

    model_config = {
        "populate_by_name": True,
    }


class PlayerIdentifierRelationships(BaseRelationships):
    """Relationships for the PlayerIdentifier model."""

    organizations: Relationship
    player: Relationship


class PlayerIdentifier(Base):
    """PlayerIdentifier model representing a player's identifier."""

    type: str = "playerIdentifier"
    attributes: PlayerIdentifierAttributes
    relationships: PlayerIdentifierRelationships
    relationships: PlayerIdentifierRelationships


class PlayerCounterAttributes(BaseModel):
    """Attributes for the PlayerCounter model."""

    name: str
    value: int


class PlayerCounterRelationships(BaseRelationships):
    """Relationships for the PlayerCounter model."""

    organization: Relationship
    player: Relationship


class PlayerCounter(Base):
    """PlayerCounter model representing a player's counter."""

    type: str = "playerCounter"
    attributes: PlayerCounterAttributes
    relationships: PlayerCounterRelationships


class PlayerStatsAttributes(BaseModel):
    """Attributes for the PlayerStats model."""

    first_time_session_duration: float = Field(alias="firstTimeSessionDuration")
    max_players: float = Field(alias="maxPlayers")
    min_players: float = Field(alias="minPlayers")
    session_duration: float = Field(alias="sessionDuration")
    unique_players: float = Field(alias="uniquePlayers")
    unique_players_by_country: float = Field(alias="uniquePlayersByCountry")

    model_config = {
        "populate_by_name": True,
    }


class PlayerStatsRelationships(BaseRelationships):
    """Relationships for the PlayerStats model."""

    game: Relationship
    organization: Relationship
    server: Relationship


class PlayerStats(Base):
    """PlayerStats model representing a player's statistics."""

    type: str = "playerStats"
    attributes: PlayerStatsAttributes
    relationships: PlayerStatsRelationships


class QuickMatchIdentifierAttributes(BaseModel):
    """Attributes for the QuickMatchIdentifier model."""

    type: IdentifierTypesLiteral  # type: ignore[reportInvalidTypeForm]
    identifier: str
    last_seen: str = Field(alias="lastSeen")
    private: bool
    metadata: object | None = None

    model_config = {
        "populate_by_name": True,
    }


class QuickMatchIdentifierRelationships(BaseRelationships):
    """Relationships for the QuickMatchIdentifier model."""

    player: Relationship
    organizations: list[Relationship] | None = None


class QuickMatchIdentifier(Base):
    """QuickMatchIdentifier model representing a quick match result.

    This is returned by the POST /players/quick-match endpoint.
    """

    type: str = "identifier"
    attributes: QuickMatchIdentifierAttributes
    relationships: QuickMatchIdentifierRelationships


class RelatedPlayerIdentifierAttributes(BaseModel):
    """Attributes for the RelatedPlayerIdentifier model."""

    identifier: str
    last_seen: str = Field(alias="lastSeen")
    metadata: object | None = None
    private: bool
    type: IdentifierTypesLiteral  # type: ignore[reportInvalidTypeForm]

    model_config = {
        "populate_by_name": True,
    }


class RelatedPlayerIdentifierRelationships(BaseRelationships):
    """Relationships for the RelatedPlayerIdentifier model."""

    organizations: Relationship
    player: Relationship
    related_identifier: Relationship = Field(alias="relatedIdentifier")
    related_players: Relationship = Field(alias="relatedPlayers")


class RelatedPlayerIdentifierMeta(BaseModel):
    """Metadata for the RelatedPlayerIdentifier model."""

    common_identifier: bool = Field(alias="commonIdentifier")

    model_config = {
        "populate_by_name": True,
    }


class RelatedPlayerIdentifier(Base):
    """RelatedPlayerIdentifier model representing a related player's identifier."""

    type: str = "relatedPlayerIdentifier"
    attributes: RelatedPlayerIdentifierAttributes
    meta: RelatedPlayerIdentifierMeta
    relationships: RelatedPlayerIdentifierRelationships



from pydantic import BaseModel

from .base import Base, BaseRelationships, Relationship


class CommandsActivityRelationships(BaseRelationships):
    """Relationships for the CommandsActivity model."""

    games: Relationship
    organization: Relationship | None = None
    user: Relationship | None = None


class CommandsActivity(Base):
    """Represents the CommandsActivity model."""

    type: str = "commandStats"
    relationships: CommandsActivityRelationships


class CoplayAttributes(BaseModel):
    """Attributes for the Coplay model."""

    duration: int
    name: str


class Coplay(Base):
    """Represents the time two players have spent together."""

    type: str = "coplay"
    attributes: CoplayAttributes


class LeaderboardAttributes(BaseModel):
    """Attributes for the Leaderboard model."""

    name: str
    rank: int
    value: int


class Leaderboard(Base):
    """Represents a leaderboard entry."""

    type: str = "leaderboardPlayer"
    attributes: LeaderboardAttributes



from __future__ import annotations

from typing import Any, Literal

from pydantic import BaseModel, Field, field_validator

__all__ = (
    "Base",
    "Relationship",
    "BaseRelationships",
    "IDENTIFIER_TYPES",
    "IdentifierTypesLiteral",
)

IDENTIFIER_TYPES = [
    "steamID",
    "BEGUID",
    "legacyBEGUID",
    "ip",
    "name",
    "survivorName",
    "steamFamilyShareOwner",
    "conanCharName",
    "egsID",
    "eosID",
    "funcomID",
    "playFabID",
    "mcUUID",
    "7dtdEOS",
    "battlebitHWID",
    "hllWindowsID",
    "palworldUID",
    "reforgerUUID",
]
IdentifierTypesLiteral = Literal[*IDENTIFIER_TYPES]


class Base(BaseModel):
    """Base model for all Battlemetrics models.

    This model follows the Json API specification, which requires a `type` and `id` field.
    """

    type: str = Field(..., description="The type of the model.")
    id: str | int = Field(..., description="The unique identifier for the model.")


class Relationship(BaseModel):
    """Represents a relationship between two resources."""

    type: str
    id: str


class BaseRelationships(BaseModel):
    """Base model for all Battlemetrics relationships."""

    @field_validator("*", mode="before")
    @classmethod
    def flatten_data(cls, v: Any) -> Any:
        """Flatten the data field if it exists."""
        if isinstance(v, dict) and "data" in v:
            return v["data"]  # type: ignore[reportUnknownVariableType]
        return v  # type: ignore[reportUnknownVariableType]



from __future__ import annotations

import asyncio
import uuid
from enum import Enum
from logging import getLogger
from typing import TYPE_CHECKING, Any, ClassVar, Literal

import aiohttp
import yarl

from .errors import BMException, Forbidden, HTTPException, NotFound, Unauthorized

if TYPE_CHECKING:
    from asyncio import AbstractEventLoop
    from types import TracebackType

    from aiohttp import BaseConnector, ClientResponse, ClientSession
    from yarl import URL


_log = getLogger(__name__)


class IdentifierType(Enum):
    """A player identifier type."""

    BE_GUID = "BEGUID"
    BE_LEGACY_GUID = "legacyBEGUID"
    CONAN_CHAR_NAME = "conanCharName"
    EGS_ID = "egsID"
    FUNCOM_ID = "funcomID"
    IP = "ip"
    MC_UUID = "mcUUID"
    NAME = "name"
    PLAY_FAB_ID = "playFabID"
    STEAM_FAMILY_SHARE_OWNER = "steamFamilyShareOwner"
    STEAM_ID = "steamID"
    SURVIVOR_NAME = "survivorName"

    def __repr__(self) -> str:
        """Return a string representation of the identifier type."""
        return f"<{self.__class__.__name__}.{self.name}>"


async def json_or_text(
    response: ClientResponse,
) -> dict[str, Any] | list[dict[str, Any]] | str:
    """
    Process a `ClientResponse` to return either a JSON object or raw text.

    This function attempts to parse the response as JSON. If the content type of the response is not
    application/json or parsing fails, it falls back to returning the raw text of the response.

    Parameters
    ----------
    response : ClientResponse
        The response object to process.

    Returns
    -------
    dict[str, t.Any] | list[dict[str, t.Any]] | str
        The parsed JSON object as a dictionary or list of dictionaries, or the raw response text.
    """
    try:
        if "application/json" in response.headers["content-type"].lower():
            return await response.json()
    except KeyError:
        # Thanks Cloudflare
        pass

    return await response.text(encoding="utf-8")


METHODS = Literal[
    "GET",
    "HEAD",
    "OPTIONS",
    "TRACE",
    "PUT",
    "DELETE",
    "POST",
    "PATCH",
    "CONNECT",
]


class Route:
    """Represents a route for the BattleMetrics API.

    This method requires either one of path or url.

    Parameters
    ----------
    method : str
        The HTTP method for the route.
    path : str
        The path for the route.
    url : str | URL
        The URL for the route.
    parameters : int | str | bool
        Optional parameters for the route.
    """

    BASE: ClassVar[str] = "https://api.battlemetrics.com"

    def __init__(
        self,
        method: METHODS,
        path: str,
        **parameters: int | str | bool,
    ) -> None:
        self.method: str = method
        url = path if path.startswith(("http://", "https://")) else f"{self.BASE}{path}"
        self.url: URL = (
            yarl.URL(url).update_query(**parameters) if parameters else yarl.URL(url)
        )


class HTTPClient:
    """Represent an HTTP Client used for making requests to APIs."""

    def __init__(
        self,
        api_key: str,
        *,
        connector: BaseConnector | None = None,
        loop: AbstractEventLoop | None = None,
        proxy: str | None = None,
        proxy_auth: aiohttp.BasicAuth | None = None,
    ) -> None:
        self.loop = loop or asyncio.get_event_loop()
        self.connector = connector
        self.proxy = proxy
        self.proxy_auth = proxy_auth

        self.__session: ClientSession = None  # type: ignore[reportAttributeAccessIssue]

        self.api_key: str = api_key

        self.ensure_session()

    def __aexit__(
        self,
        exc_type: BaseException | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        """Close the HTTP client when exiting."""
        return self.close()

    def ensure_session(self) -> None:
        """
        Ensure that an :class:`ClientSession` is created and open.

        If a session does not exist, this method creates a new :class:`ClientSession`
        using the provided connector and loop.
        """
        if not self.__session or self.__session.closed:
            self.__session = aiohttp.ClientSession(
                connector=self.connector,
                loop=self.loop,
            )

    async def close(self) -> None:
        """Close the :class:`ClientSession` if it exists and is open."""
        if self.__session:
            await self.__session.close()

    async def request(
        self,
        route: Route,
        **kwargs: Any,
    ) -> Any:
        """
        Send a request to the specified route and returns the response.

        This method constructs and sends an HTTP request based on the specified route and headers.
        It processes the response to return JSON data or raw text, handling errors as needed.

        Parameters
        ----------
        route : Route
            The route object containing the method and URL for the request.

        Returns
        -------
        dict[str, t.Any] | list[dict[str, t.Any]] | str
            The response data as a parsed JSON object or list, or raw text if JSON parsing is
            not applicable.

        Raises
        ------
        BMException
            Will raise if the request fails or the response indicates an error.
            Might raise a more specific exception if the response status code is known.
        """
        self.ensure_session()

        method = route.method
        url = route.url
        path = route.url.path

        headers: dict[str, Any] = {
            "Accept": "application/json",
            **(kwargs.get("headers") or {}),
        }

        if self.api_key:
            headers["Authorization"] = f"Bearer {self.api_key}"
        if "json" in kwargs:
            headers["Content-Type"] = "application/json"

        kwargs["headers"] = headers

        if self.proxy:
            kwargs["proxy"] = self.proxy
        if self.proxy_auth:
            kwargs["proxy_auth"] = self.proxy_auth

        async with self.__session.request(method, url, **kwargs) as response:
            _log.debug("%s %s returned %s", method, path, response.status)

            # errors typically have text involved, so this should be safe 99.5% of the time.
            data = await json_or_text(response)

            if 200 <= response.status < 300:
                return data

            if isinstance(data, dict):
                if response.status == 401:
                    _log.warning(
                        "Path %s returned 401, your API key may be invalid.",
                        path,
                    )
                    raise Unauthorized(response, data)
                if response.status == 403:
                    _log.warning(
                        "Path %s returned 403, check whether you have valid permissions.",
                        path,
                    )
                    raise Forbidden(response, data)
                if response.status == 404:
                    _log.warning(
                        "Path %s returned 404, check whether the path is correct.",
                        path,
                    )
                    raise NotFound(response, data)
                if response.status == 429:
                    _log.warning(
                        "We're being rate limited. You are limited to %s requests per minute.",
                        response.headers.get("X-Rate-Limit-Limit"),
                    )

                raise HTTPException(response, data)

            raise BMException

    # HTTP Requests

    async def create_ban(
        self,
        player_id: int,
        *,
        reason: str,
        note: str | None,
        banlist_id: str,
        organization_id: int,
        server_id: int,
        org_wide: bool = True,
        auto_add_enabled: bool = True,
        native_enabled: bool = True,
        identifiers: list[str | dict[str, Any]] | None = None,
        expires: str | None = None,
    ) -> dict[str, Any]:
        """Create a ban with all required and optional parameters."""
        data: dict[str, Any] = {
            "data": {
                "type": "ban",
                "attributes": {
                    "uid": str(uuid.uuid4())[:14],
                    "reason": reason,
                    "note": note,
                    "expires": expires,
                    "identifiers": identifiers or [],  # TODO: Add identifier handling
                    "orgWide": org_wide,
                    "autoAddEnabled": auto_add_enabled,
                    "nativeEnabled": native_enabled,
                },
                "relationships": {
                    "organization": {
                        "data": {
                            "type": "organization",
                            "id": f"{organization_id}",
                        },
                    },
                    "server": {
                        "data": {
                            "type": "server",
                            "id": f"{server_id}",
                        },
                    },
                    "player": {
                        "data": {
                            "type": "player",
                            "id": f"{player_id}",
                        },
                    },
                    "banList": {
                        "data": {
                            "type": "banList",
                            "id": f"{banlist_id}",
                        },
                    },
                },
            },
        }

        return await self.request(Route(method="POST", path="/bans"), json=data)

    async def import_bans(self, bans: list[dict[str, Any]]) -> list[dict[str, Any]]:
        """Import multiple bans at once.

        This method allows adding multiple bans at once with certain limitations.

        Parameters
        ----------
        bans : list[dict[str, Any]]
            A list of ban data dictionaries, each containing attributes and relationships.

        Returns
        -------
        Any
            The response from the API.

        Raises
        ------
        BMException
            Will raise if the request fails or the response indicates an error.
        """
        data = {"data": bans}

        return await self.request(Route(method="POST", path="/bans/import"), json=data)

    async def export_bans(
        self,
        file_format: Literal[
            "arma2/bans.txt",
            "arma3/bans.txt",
            "squad/Bans.cfg",
            "ark/banlist.txt",
            "rust/bans.cfg",
            "rust/bansip_SERVER.ini",
        ],
        *,
        organization_id: int | None = None,
        server_id: int | None = None,
    ) -> Any:
        """Export bans in a specific format.

        Parameters
        ----------
        file_format : str
            The format to export the bans in.

        Returns
        -------
        Any
            The response from the API containing the exported bans.

        Raises
        ------
        BMException
            Will raise if the request fails or the response indicates an error.
        """
        params = {"format": file_format}
        if organization_id:
            params["filter[organization]"] = str(organization_id)
        if server_id:
            params["filter[server]"] = str(server_id)

        return await self.request(
            Route(method="GET", path="/bans/export"),
            params=params,
        )

    async def delete_ban(self, ban_id: int) -> None:
        """Delete a specific ban by its ID.

        Parameters
        ----------
        ban_id : int
            The ID of the ban to delete.


        Raises
        ------
        BMException
            Will raise if the request fails or the response indicates an error.
        """
        return await self.request(Route(method="DELETE", path=f"/bans/{ban_id}"))

    async def ban_info(self, ban_id: int) -> dict[str, Any]:
        """Get information about a specific ban."""
        return await self.request(
            Route(
                method="GET",
                path=f"/bans/{ban_id}",
                include="organization,player,server,user",
            ),
        )

    async def list_bans(
        self,
        *,
        banlist_id: str | None = None,
        exempt: bool | None = None,
        expired: bool | None = None,
        organization_id: int | None = None,
        player_id: int | None = None,
        search: str | None = None,
        server_id: int | None = None,
        user_ids: list[int] | None = None,
        page_size: int | None = None,
    ) -> dict[str, Any]:
        """List bans with optional filters for organization and server.

        Parameters
        ----------
        organization_id : str | None
            The ID of the organization to filter by.
        server_id : str | None
            The ID of the server to filter by.
        page : int
            The page number for pagination.
        per_page : int
            The number of results per page.

        Returns
        -------
        Any
            The response from the API containing the list of bans.

        Raises
        ------
        BMException
            Will raise if the request fails or the response indicates an error.
        """
        params: dict[str, Any] = {
            "include": "organization,player,server,user",
        }

        if banlist_id:
            params["filter[banList]"] = banlist_id
        if exempt is not None:
            params["filter[exempt]"] = str(exempt).lower()
        if expired is not None:
            params["filter[expired]"] = str(expired).lower()
        if player_id:
            params["filter[player]"] = player_id
        if search:
            params["filter[search]"] = search
        if user_ids:
            params["filter[users]"] = ",".join(map(str, user_ids))
        if organization_id:
            params["filter[organization]"] = organization_id
        if server_id:
            params["filter[server]"] = server_id
        if page_size:
            params["page[size]"] = page_size

        return await self.request(Route(method="GET", path="/bans"), params=params)

    async def update_ban(
        self,
        ban_id: int | None = None,
        *,
        reason: str | None = None,
        note: str | None = None,
        identifiers: list[str | dict[str, Any]] | None = None,
        expires: str | None = None,
        org_wide: bool = True,
        auto_add_enabled: bool = True,
        native_enabled: bool = True,
    ) -> dict[str, Any]:
        """Update a specific ban by its ID.

        Parameters
        ----------
        ban_id : int
            The ID of the ban to update.

        Raises
        ------
        BMException
            Will raise if the request fails or the response indicates an error.
        """
        data: dict[str, Any] = {
            "data": {
                "type": "ban",
                "attributes": {
                    "reason": reason,
                    "note": note,
                    "expires": expires,
                    "identifiers": identifiers or [],  # TODO: Add identifier handling
                    "orgWide": org_wide,
                    "autoAddEnabled": auto_add_enabled,
                    "nativeEnabled": native_enabled,
                },
            },
        }
        return await self.request(
            Route(method="PATCH", path=f"/bans/{ban_id}"),
            json=data,
        )

    async def create_banlist_exemption(
        self,
        ban_id: int,
        organization_id: int,
        *,
        reason: str | None = None,
    ) -> dict[str, Any]:
        """Create a banlist Exemption."""
        data = {
            "data": {
                "type": "banExemption",
                "attributes": {"reason": reason},
                "relationships": {
                    "organization": {
                        "data": {"type": "organization", "id": str(organization_id)},
                    },
                },
            },
        }
        return await self.request(
            Route(method="POST", path=f"/bans/{ban_id}/relationships/exemptions"),
            json=data,
        )

    async def read_banlist_exemption(
        self,
        ban_id: int,
        ban_exemption_id: int,
    ) -> dict[str, Any]:
        """Read a specific banlist exemption by its ID.

        Parameters
        ----------
        ban_id : int
            The ID of the ban to read.
        ban_exemption_id : int
            The ID of the ban exemption to read.

        Returns
        -------
        dict[str, Any]
            The response from the API containing the banlist exemption.

        Raises
        ------
        BMException
            Will raise if the request fails or the response indicates an error.
        """
        return await self.request(
            Route(
                method="GET",
                path=f"/bans/{ban_id}/relationships/exemptions/{ban_exemption_id}",
            ),
        )

    async def list_banlist_exemptions(self, ban_id: int) -> dict[str, Any]:
        """List all banlist exemptions for a specific ban.

        Parameters
        ----------
        ban_id : int
            The ID of the ban to list exemptions for.

        Returns
        -------
        dict[str, Any]
            The response from the API containing the list of banlist exemptions.

        Raises
        ------
        BMException
            Will raise if the request fails or the response indicates an error.
        """
        return await self.request(
            Route(method="GET", path=f"/bans/{ban_id}/relationships/exemptions"),
        )

    async def update_banlist_exemption(
        self,
        ban_id: int,
        *,
        reason: str | None = None,
    ) -> dict[str, Any]:
        """Update a specific banlist exemption by its ID.

        Parameters
        ----------
        ban_id : int
            The ID of the ban to update.
        ban_exemption_id : int
            The ID of the ban exemption to update.
        reason : str | None
            The new reason for the ban exemption.

        Returns
        -------
        dict[str, Any]
            The response from the API containing the updated banlist exemption.

        Raises
        ------
        BMException
            Will raise if the request fails or the response indicates an error.
        """
        data = {
            "data": {
                "type": "banExemption",
                "attributes": {"reason": reason},
            },
        }
        return await self.request(
            Route(method="PATCH", path=f"/bans/{ban_id}/relationships/exemptions"),
            json=data,
        )

    async def delete_banlist_exemption(
        self,
        ban_id: int,
    ) -> None:
        """Delete a specific banlist exemption by its ID.

        Parameters
        ----------
        ban_id : int
            The ID of the ban to delete the exemption from.
        ban_exemption_id : int
            The ID of the ban exemption to delete.

        Returns
        -------
        dict[str, Any]
            The response from the API confirming the deletion.

        Raises
        ------
        BMException
            Will raise if the request fails or the response indicates an error.
        """
        await self.request(
            Route(method="DELETE", path=f"/bans/{ban_id}/relationships/exemptions"),
        )

    # ----------------------------- Ban Lists ----------------------------- #

    async def create_banlist(
        self,
        *,
        name: str,
        action: Literal["none", "log", "kick"],
        organization_id: int,
        owner_id: int,
        default_identifiers: list[str] | None = None,
        default_reasons: list[str] | None = None,
        default_auto_add_enabled: bool = True,
        default_native_enabled: bool | None = None,
        native_ban_ttl: int | None = None,
        native_ban_temp_max_expires: int | None = None,
        native_ban_perm_max_expires: int | None = None,
    ) -> dict[str, Any]:
        """Create a ban list."""
        data = {
            "data": {
                "type": "banList",
                "attributes": {
                    "name": name,
                    "action": action,
                    "defaultIdentifiers": default_identifiers or [],
                    "defaultReasons": default_reasons or [],
                    "defaultAutoAddEnabled": default_auto_add_enabled,
                    "defaultNativeEnabled": default_native_enabled,
                    "nativeBanTTL": native_ban_ttl,
                    "nativeBanTempMaxExpires": native_ban_temp_max_expires,
                    "nativeBanPermMaxExpires": native_ban_perm_max_expires,
                },
                "relationships": {
                    "organization": {
                        "data": {"type": "organization", "id": str(organization_id)},
                    },
                    "owner": {"data": {"type": "organization", "id": str(owner_id)}},
                },
            },
        }
        return await self.request(Route(method="POST", path="/ban-lists"), json=data)

    async def create_banlist_from_invite(
        self,
        *,
        code: str,
        action: Literal["none", "log", "kick"],
        organization_id: int,
        owner_id: int,
        default_identifiers: list[str] | None = None,
        default_reasons: list[str] | None = None,
        default_auto_add_enabled: bool = True,
        default_native_enabled: bool | None = None,
        native_ban_ttl: int | None = None,
        native_ban_temp_max_expires: int | None = None,
        native_ban_perm_max_expires: int | None = None,
    ) -> dict[str, Any]:
        """Accept a ban list invite and create a subscription."""
        data = {
            "data": {
                "type": "banList",
                "attributes": {
                    "code": code,
                    "action": action,
                    "defaultIdentifiers": default_identifiers or [],
                    "defaultReasons": default_reasons or [],
                    "defaultAutoAddEnabled": default_auto_add_enabled,
                    "defaultNativeEnabled": default_native_enabled,
                    "nativeBanTTL": native_ban_ttl,
                    "nativeBanTempMaxExpires": native_ban_temp_max_expires,
                    "nativeBanPermMaxExpires": native_ban_perm_max_expires,
                },
                "relationships": {
                    "organization": {
                        "data": {"type": "organization", "id": str(organization_id)},
                    },
                    "owner": {"data": {"type": "organization", "id": str(owner_id)}},
                },
            },
        }
        return await self.request(
            Route(method="POST", path="/ban-lists/accept-invite"),
            json=data,
        )

    async def list_banlists(
        self,
        *,
        include: str | None = None,
        page_size: int | None = None,
    ) -> dict[str, Any]:
        """List ban lists you own or are subscribed to."""
        params: dict[str, Any] = {}
        if include:
            params["include"] = include
        if page_size:
            params["page[size]"] = page_size
        return await self.request(Route(method="GET", path="/ban-lists"), params=params)

    async def get_banlist(
        self,
        banlist_id: str,
        *,
        include: str | None = None,
    ) -> dict[str, Any]:
        """Get a specific ban list."""
        params: dict[str, Any] = {}
        if include:
            params["include"] = include
        return await self.request(
            Route(method="GET", path=f"/ban-lists/{banlist_id}"),
            params=params,
        )

    async def update_banlist(
        self,
        banlist_id: str,
        *,
        name: str | None = None,
        action: Literal["none", "log", "kick"] | None = None,
        default_identifiers: list[str] | None = None,
        default_reasons: list[str] | None = None,
        default_auto_add_enabled: bool | None = None,
        default_native_enabled: bool | None = None,
        native_ban_ttl: int | None = None,
        native_ban_temp_max_expires: int | None = None,
        native_ban_perm_max_expires: int | None = None,
    ) -> dict[str, Any]:
        """Update a ban list."""
        attributes = {
            "name": name,
            "action": action,
            "defaultIdentifiers": default_identifiers,
            "defaultReasons": default_reasons,
            "defaultAutoAddEnabled": default_auto_add_enabled,
            "defaultNativeEnabled": default_native_enabled,
            "nativeBanTTL": native_ban_ttl,
            "nativeBanTempMaxExpires": native_ban_temp_max_expires,
            "nativeBanPermMaxExpires": native_ban_perm_max_expires,
        }
        # remove None values
        attributes = {k: v for k, v in attributes.items() if v is not None}
        data = {"data": {"type": "banList", "id": banlist_id, "attributes": attributes}}
        return await self.request(
            Route(method="PATCH", path=f"/ban-lists/{banlist_id}"),
            json=data,
        )

    async def list_banlist_organizations(
        self,
        banlist_id: str,
        *,
        include: str | None = None,
        page_size: int | None = None,
    ) -> dict[str, Any]:
        """List organizations subscribed to a ban list."""
        params: dict[str, Any] = {}
        if include:
            params["include"] = include
        if page_size:
            params["page[size]"] = page_size
        return await self.request(
            Route(
                method="GET",
                path=f"/ban-lists/{banlist_id}/relationships/organizations",
            ),
            params=params,
        )

    async def get_banlist_subscription(
        self,
        banlist_id: str,
        organization_id: int,
        *,
        include: str | None = None,
    ) -> dict[str, Any]:
        """Read an organization's subscription to a ban list."""
        params: dict[str, Any] = {}
        if include:
            params["include"] = include
        return await self.request(
            Route(
                method="GET",
                path=f"/ban-lists/{banlist_id}/relationships/organizations/{organization_id}",
            ),
            params=params,
        )

    async def update_banlist_subscription(
        self,
        banlist_id: str,
        organization_id: int,
        *,
        perm_manage: bool | None = None,
        perm_create: bool | None = None,
        perm_update: bool | None = None,
        perm_delete: bool | None = None,
    ) -> dict[str, Any]:
        """Update an organization's subscription permissions for a ban list."""
        attributes = {
            "permManage": perm_manage,
            "permCreate": perm_create,
            "permUpdate": perm_update,
            "permDelete": perm_delete,
        }
        attributes = {k: v for k, v in attributes.items() if v is not None}
        data = {"data": {"type": "banList", "id": banlist_id, "attributes": attributes}}
        return await self.request(
            Route(
                method="PATCH",
                path=f"/ban-lists/{banlist_id}/relationships/organizations/{organization_id}",
            ),
            json=data,
        )

    async def unsubscribe_banlist(self, banlist_id: str, organization_id: int) -> None:
        """Unsubscribe an organization from a ban list."""
        await self.request(
            Route(
                method="DELETE",
                path=f"/ban-lists/{banlist_id}/relationships/organizations/{organization_id}",
            ),
        )

    # -------------------------- Ban List Invites ------------------------- #

    async def create_banlist_invite(
        self,
        banlist_id: str,
        *,
        limit: int | None = None,
        perm_manage: bool | None = None,
        perm_create: bool | None = None,
        perm_update: bool | None = None,
        perm_delete: bool | None = None,
        uses: int | None = None,
        organization_id: int | None = None,
    ) -> dict[str, Any]:
        """Create a ban list invite."""
        attributes = {
            "limit": limit,
            "permManage": perm_manage,
            "permCreate": perm_create,
            "permUpdate": perm_update,
            "permDelete": perm_delete,
            "uses": uses,
        }
        attributes = {k: v for k, v in attributes.items() if v is not None}
        relationships: dict[str, Any] = {}
        if organization_id is not None:
            relationships["organization"] = {
                "data": {"type": "organization", "id": str(organization_id)},
            }
        data = {
            "data": {
                "type": "banListInvite",
                "attributes": attributes,
                "relationships": relationships,
            },
        }
        return await self.request(
            Route(method="POST", path=f"/ban-lists/{banlist_id}/relationships/invites"),
            json=data,
        )

    async def list_banlist_invites(
        self,
        banlist_id: str,
        *,
        include: str | None = None,
    ) -> dict[str, Any]:
        """List invites for a ban list."""
        params: dict[str, Any] = {}
        if include:
            params["include"] = include
        return await self.request(
            Route(method="GET", path=f"/ban-lists/{banlist_id}/relationships/invites"),
            params=params,
        )

    async def get_banlist_invite(
        self,
        invite_id: str,
        *,
        include: str | None = None,
    ) -> dict[str, Any]:
        """Get a specific ban list invite."""
        params: dict[str, Any] = {}
        if include:
            params["include"] = include
        return await self.request(
            Route(method="GET", path=f"/ban-list-invites/{invite_id}"),
            params=params,
        )

    async def delete_banlist_invite(self, invite_id: str) -> None:
        """Delete a ban list invite."""
        await self.request(
            Route(method="DELETE", path=f"/ban-list-invites/{invite_id}"),
        )

    # ----------------------------- Native Bans ---------------------------- #

    async def list_native_bans(
        self,
        *,
        ban_id: int | None = None,
        server_id: int | None = None,
        include: str | None = None,
        page_size: int | None = None,
        sort: str | None = None,
    ) -> dict[str, Any]:
        """List native bans (game specific)."""
        params: dict[str, Any] = {}
        if ban_id:
            params["filter[ban]"] = ban_id
        if server_id:
            params["filter[server]"] = server_id
        if include:
            params["include"] = include
        if page_size:
            params["page[size]"] = page_size
        if sort:
            params["sort"] = sort
        return await self.request(
            Route(method="GET", path="/bans-native"),
            params=params,
        )

    async def force_update_native_ban(self, native_ban_id: str) -> dict[str, Any]:
        """Force update a native ban."""
        return await self.request(
            Route(method="POST", path=f"/bans-native/{native_ban_id}/force-update"),
        )

    # --------------------------- Commands Activity ------------------------ #

    async def list_command_stats(
        self,
        organization_id: int,
        *,
        commands: list[str] | None = None,
        servers: list[int] | None = None,
        users: list[int] | None = None,
        timestamp_range: str | None = None,
        summary: bool | None = None,
    ) -> dict[str, Any]:
        """List command activity statistics for an organization."""
        params: dict[str, Any] = {}
        if commands:
            params["filter[commands]"] = ",".join(commands)
        if servers:
            params["filter[servers]"] = ",".join(map(str, servers))
        if users:
            params["filter[users]"] = ",".join(map(str, users))
        if timestamp_range:
            params["filter[timestamp]"] = timestamp_range
        if summary is not None:
            params["filter[summary]"] = str(summary).lower()
        return await self.request(
            Route(
                method="GET",
                path=f"/organizations/{organization_id}/relationships/command-stats",
            ),
            params=params,
        )

    # -------------------------------- Coplay ------------------------------ #

    async def list_coplay(
        self,
        player_id: int,
        *,
        period: str,
        players: list[int] | None = None,
        servers: list[int] | None = None,
        organizations: list[int] | None = None,
        page_size: int | None = None,
    ) -> dict[str, Any]:
        """List coplay data for a player."""
        params: dict[str, Any] = {"filter[period]": period}
        if players:
            params["filter[players]"] = ",".join(map(str, players))
        if servers:
            params["filter[servers]"] = ",".join(map(str, servers))
        if organizations:
            params["filter[organizations]"] = ",".join(map(str, organizations))
        if page_size:
            params["page[size]"] = page_size
        return await self.request(
            Route(method="GET", path=f"/players/{player_id}/relationships/coplay"),
            params=params,
        )

    # ------------------------------ Data Points --------------------------- #

    async def get_metrics(
        self,
        metrics: list[dict[str, Any]],
        *,
        fields: list[str] | None = None,
    ) -> dict[str, Any]:
        """Get data points for metrics. Each metric dict should include name, range, resolution."""
        params: dict[str, Any] = {}
        if fields:
            params["fields[dataPoint]"] = ",".join(fields)
        # Only first metric supported in this convenience wrapper; call multiple times for more.
        for metric in metrics:
            for key, value in metric.items():
                params[f"metrics[{key}]"] = value
            break  # only first metric for now (API supports multiple via different encoding)
        return await self.request(Route(method="GET", path="/metrics"), params=params)

    # ------------------------------ Player Flags -------------------------- #

    async def create_player_flag(
        self,
        *,
        organization_id: int,
        name: str,
        color: str | None = None,
        icon: str | None = None,
        description: str | None = None,
    ) -> dict[str, Any]:
        """Create a player flag."""
        data = {
            "data": {
                "type": "playerFlag",
                "attributes": {
                    "name": name,
                    "color": color,
                    "icon": icon,
                    "description": description,
                },
                "relationships": {
                    "organization": {
                        "data": {"type": "organization", "id": str(organization_id)},
                    },
                },
            },
        }
        return await self.request(Route(method="POST", path="/player-flags"), json=data)

    async def get_player_flag(
        self,
        flag_id: str,
        *,
        include: str | None = None,
    ) -> dict[str, Any]:
        """Get a specific player flag."""
        params: dict[str, Any] = {}
        if include:
            params["include"] = include
        return await self.request(
            Route(method="GET", path=f"/player-flags/{flag_id}"),
            params=params,
        )

    async def list_player_flags(
        self,
        *,
        personal: bool | None = None,
        include: str | None = None,
        page_size: int | None = None,
    ) -> dict[str, Any]:
        """List player flags."""
        params: dict[str, Any] = {}
        if personal is not None:
            params["filter[personal]"] = str(personal).lower()
        if include:
            params["include"] = include
        if page_size:
            params["page[size]"] = page_size
        return await self.request(
            Route(method="GET", path="/player-flags"),
            params=params,
        )

    async def update_player_flag(
        self,
        flag_id: str,
        *,
        name: str | None = None,
        color: str | None = None,
        icon: str | None = None,
        description: str | None = None,
    ) -> dict[str, Any]:
        """Update a player flag."""
        attributes = {
            "name": name,
            "color": color,
            "icon": icon,
            "description": description,
        }
        attributes = {k: v for k, v in attributes.items() if v is not None}
        data = {"data": {"type": "playerFlag", "id": flag_id, "attributes": attributes}}
        return await self.request(
            Route(method="PATCH", path=f"/player-flags/{flag_id}"),
            json=data,
        )

    async def delete_player_flag(self, flag_id: str) -> None:
        """Delete a player flag."""
        await self.request(Route(method="DELETE", path=f"/player-flags/{flag_id}"))

    async def list_player_flag_assignments(
        self,
        player_id: int,
        *,
        include: str | None = None,
        page_size: int | None = None,
    ) -> dict[str, Any]:
        """List flags applied to a player."""
        params: dict[str, Any] = {}
        if include:
            params["include"] = include
        if page_size:
            params["page[size]"] = page_size
        return await self.request(
            Route(method="GET", path=f"/players/{player_id}/relationships/flags"),
            params=params,
        )

    async def add_player_flag_assignment(
        self,
        player_id: int,
        player_flag_id: str,
    ) -> dict[str, Any]:
        """Add a flag to a player."""
        data = {
            "data": {
                "type": "flagPlayer",
                "relationships": {
                    "playerFlag": {
                        "data": {"type": "playerFlag", "id": player_flag_id},
                    },
                },
            },
        }
        return await self.request(
            Route(method="POST", path=f"/players/{player_id}/relationships/flags"),
            json=data,
        )

    async def remove_player_flag_assignment(
        self,
        player_id: int,
        player_flag_id: str,
    ) -> None:
        """Remove a flag from a player."""
        await self.request(
            Route(
                method="DELETE",
                path=f"/players/{player_id}/relationships/flags/{player_flag_id}",
            ),
        )

    # --------------------------------- Games ------------------------------ #

    async def list_games(self, *, page_size: int | None = None) -> dict[str, Any]:
        """List supported games."""
        params: dict[str, Any] = {}
        if page_size:
            params["page[size]"] = page_size
        return await self.request(Route(method="GET", path="/games"), params=params)

    async def get_game(self, game_id: str) -> dict[str, Any]:
        """Get information about a game."""
        return await self.request(Route(method="GET", path=f"/games/{game_id}"))

    # ---------------------------- Game Features --------------------------- #

    async def list_game_features(
        self,
        *,
        game: str | None = None,
        page_size: int | None = None,
    ) -> dict[str, Any]:
        """List game features (filter by game optionally)."""
        params: dict[str, Any] = {}
        if game:
            params["filter[game]"] = game
        if page_size:
            params["page[size]"] = page_size
        return await self.request(
            Route(method="GET", path="/game-features"),
            params=params,
        )

    async def list_game_feature_options(
        self,
        feature_id: str,
        *,
        page_size: int | None = None,
        sort: str | None = None,
    ) -> dict[str, Any]:
        """List options for a given game feature."""
        params: dict[str, Any] = {}
        if page_size:
            params["page[size]"] = page_size
        if sort:
            params["sort"] = sort
        return await self.request(
            Route(
                method="GET",
                path=f"/game-features/{feature_id}/relationships/options",
            ),
            params=params,
        )

    # ------------------------------ Leaderboards ------------------------- #

    async def get_leaderboard(
        self,
        server_id: int,
        *,
        period: str,
        player_id: int | None = None,
        leaderboard: str = "time",
        page_size: int | None = None,
        offset: int | None = None,
    ) -> dict[str, Any]:
        """Get leaderboard data for a server (e.g. time)."""
        params: dict[str, Any] = {"filter[period]": period}
        if player_id is not None:
            params["filter[player]"] = player_id
        if page_size:
            params["page[size]"] = page_size
        if offset is not None:
            params["page[offset]"] = offset
        return await self.request(
            Route(
                method="GET",
                path=f"/servers/{server_id}/relationships/leaderboards/{leaderboard}",
            ),
            params=params,
        )

    # ---------------------------- Organization Stats --------------------- #

    async def get_organization_player_stats(
        self,
        organization_id: int,
        *,
        range_: str,
        game: str | None = None,
        group: str | None = None,
    ) -> dict[str, Any]:
        """Get organization player stats."""
        params: dict[str, Any] = {"filter[range]": range_}
        if game:
            params["filter[game]"] = game
        if group:
            params["filter[group]"] = group
        return await self.request(
            Route(method="GET", path=f"/organizations/{organization_id}/stats/players"),
            params=params,
        )

    # -------------------------------- Players ----------------------------- #

    async def list_players(
        self,
        *,
        search: str | None = None,
        servers: list[int] | None = None,
        organizations: list[int] | None = None,
        game: str | None = None,
        page_size: int | None = None,
        sort: str | None = None,
        online: bool | None = None,
    ) -> dict[str, Any]:
        """List/search players."""
        params: dict[str, Any] = {}
        if search:
            params["filter[search]"] = search
        if servers:
            params["filter[servers]"] = ",".join(map(str, servers))
        if organizations:
            params["filter[organizations]"] = ",".join(map(str, organizations))
        if game:
            params["filter[server][game]"] = game
        if page_size:
            params["page[size]"] = page_size
        if sort:
            params["sort"] = sort
        if online is not None:
            params["filter[online]"] = str(online).lower()
        return await self.request(Route(method="GET", path="/players"), params=params)

    async def get_player(
        self,
        player_id: int,
        *,
        include: str | None = None,
    ) -> dict[str, Any]:
        """Get player information."""
        params: dict[str, Any] = {}
        if include:
            params["include"] = include
        return await self.request(
            Route(method="GET", path=f"/players/{player_id}"),
            params=params,
        )

    async def match_players(self, identifiers: list[dict[str, str]]) -> dict[str, Any]:
        """Match players by identifiers (slow full match)."""
        data = {
            "data": [
                {"type": "identifier", "attributes": ident} for ident in identifiers
            ],
        }
        return await self.request(
            Route(method="POST", path="/players/match"),
            json=data,
        )

    async def quick_match_players(
        self,
        identifiers: list[dict[str, str]],
    ) -> dict[str, Any]:
        """Quick match players by identifiers."""
        data = {
            "data": [
                {"type": "identifier", "attributes": ident} for ident in identifiers
            ],
        }
        return await self.request(
            Route(method="POST", path="/players/quick-match"),
            json=data,
        )

    async def player_time_played_history(
        self,
        player_id: int,
        server_id: int,
        *,
        start: str,
        stop: str,
    ) -> dict[str, Any]:
        """Get a player's time played history for a server."""
        params = {"start": start, "stop": stop}
        return await self.request(
            Route(
                method="GET",
                path=f"/players/{player_id}/time-played-history/{server_id}",
            ),
            params=params,
        )

    async def get_player_server_info(
        self,
        player_id: int,
        server_id: int,
    ) -> dict[str, Any]:
        """Get player information for a specific server."""
        return await self.request(
            Route(method="GET", path=f"/players/{player_id}/servers/{server_id}"),
        )

    async def player_session_history(
        self,
        player_id: int,
        *,
        servers: list[int] | None = None,
        organizations: list[int] | None = None,
        page_size: int | None = None,
        include: str | None = None,
    ) -> dict[str, Any]:
        """Get a player's session history."""
        params: dict[str, Any] = {}
        if servers:
            params["filter[servers]"] = ",".join(map(str, servers))
        if organizations:
            params["filter[organizations]"] = ",".join(map(str, organizations))
        if page_size:
            params["page[size]"] = page_size
        if include:
            params["include"] = include
        return await self.request(
            Route(method="GET", path=f"/players/{player_id}/relationships/sessions"),
            params=params,
        )

    async def related_identifiers(
        self,
        player_id: int,
        *,
        match_identifiers: list[str] | None = None,
        identifiers: list[str] | None = None,
        page_size: int | None = None,
        include: str | None = None,
    ) -> dict[str, Any]:
        """Get related identifiers for a player."""
        params: dict[str, Any] = {}
        if match_identifiers:
            params["filter[matchIdentifiers]"] = ",".join(match_identifiers)
        if identifiers:
            params["filter[identifiers]"] = ",".join(identifiers)
        if page_size:
            params["page[size]"] = page_size
        if include:
            params["include"] = include
        return await self.request(
            Route(
                method="GET",
                path=f"/players/{player_id}/relationships/related-identifiers",
            ),
            params=params,
        )

    # ----------------------------- Player Queries ------------------------- #

    async def list_player_queries(
        self,
        *,
        organizations: list[int] | None = None,
        ids: list[str] | None = None,
        sort: str | None = None,
        page_size: int | None = None,
    ) -> dict[str, Any]:
        """List related player queries."""
        params: dict[str, Any] = {}
        if organizations:
            params["filter[organizations]"] = ",".join(map(str, organizations))
        if ids:
            params["filter[ids]"] = ",".join(ids)
        if sort:
            params["sort"] = sort
        if page_size:
            params["page[size]"] = page_size
        return await self.request(
            Route(method="GET", path="/player-queries"),
            params=params,
        )

    async def get_player_query(self, query_id: str) -> dict[str, Any]:
        """Get a related player query."""
        return await self.request(
            Route(method="GET", path=f"/player-queries/{query_id}"),
        )

    async def create_player_query(
        self,
        organization_id: int,
        *,
        query_name: str,
        conditions: list[dict[str, Any]],
    ) -> dict[str, Any]:
        """Create a related player query."""
        data = {
            "data": {
                "type": "playerQuery",
                "attributes": {"queryName": query_name, "conditions": conditions},
                "relationships": {
                    "organization": {
                        "data": {"type": "organization", "id": str(organization_id)},
                    },
                },
            },
        }
        return await self.request(
            Route(method="POST", path="/player-queries"),
            json=data,
        )

    async def update_player_query(
        self,
        query_id: str,
        *,
        query_name: str | None = None,
        conditions: list[dict[str, Any]] | None = None,
    ) -> dict[str, Any]:
        """Update a related player query."""
        attributes = {"queryName": query_name, "conditions": conditions}
        attributes = {k: v for k, v in attributes.items() if v is not None}
        data = {
            "data": {"type": "playerQuery", "id": query_id, "attributes": attributes},
        }
        return await self.request(
            Route(method="PATCH", path=f"/player-queries/{query_id}"),
            json=data,
        )

    async def delete_player_query(self, query_id: str) -> None:
        """Delete a saved related player query."""
        await self.request(Route(method="DELETE", path=f"/player-queries/{query_id}"))

    async def run_player_query(
        self,
        player_id: int,
        query_id: str,
        *,
        identifiers: list[str] | None = None,
        include: str | None = None,
        page_size: int | None = None,
        offset: int | None = None,
    ) -> dict[str, Any]:
        """Run a saved related player query for a player."""
        params: dict[str, Any] = {}
        if identifiers:
            params["filter[identifiers]"] = ",".join(identifiers)
        if include:
            params["include"] = include
        if page_size:
            params["page[size]"] = page_size
        if offset is not None:
            params["page[offset]"] = offset
        return await self.request(
            Route(
                method="GET",
                path=f"/players/{player_id}/relationships/player-query/{query_id}",
            ),
            params=params,
        )

    async def run_custom_player_query(
        self,
        player_id: int,
        *,
        conditions: list[dict[str, Any]],
        identifiers: list[str] | None = None,
        include: str | None = None,
        page_size: int | None = None,
        offset: int | None = None,
    ) -> dict[str, Any]:
        """Run a custom related player query (not saved) for a player."""
        params: dict[str, Any] = {}
        if identifiers:
            params["filter[identifiers]"] = ",".join(identifiers)
        if include:
            params["include"] = include
        if page_size:
            params["page[size]"] = page_size
        if offset is not None:
            params["page[offset]"] = offset
        data = {
            "data": {
                "type": "playerQuery",
                "attributes": {"conditions": conditions},
            },
        }
        return await self.request(
            Route(
                method="POST",
                path=f"/players/{player_id}/relationships/player-query",
            ),
            params=params,
            json=data,
        )

    # ----------------------------- Reserved Slots ------------------------- #

    async def create_reserved_slot(
        self,
        *,
        player_id: int,
        organization_id: int,
        server_ids: list[int],
        identifiers: list[str | dict[str, Any]] | None = None,
        expires: str | None = None,
    ) -> dict[str, Any]:
        """Create a reserved slot."""
        data = {
            "data": {
                "type": "reservedSlot",
                "attributes": {"expires": expires, "identifiers": identifiers or []},
                "relationships": {
                    "player": {"data": {"type": "player", "id": str(player_id)}},
                    "servers": {
                        "data": [
                            {"type": "server", "id": str(sid)} for sid in server_ids
                        ],
                    },
                    "organization": {
                        "data": {"type": "organization", "id": str(organization_id)},
                    },
                },
            },
        }
        return await self.request(
            Route(method="POST", path="/reserved-slots"),
            json=data,
        )

    async def get_reserved_slot(
        self,
        slot_id: str,
        *,
        include: str | None = None,
    ) -> dict[str, Any]:
        """Get a reserved slot."""
        params: dict[str, Any] = {}
        if include:
            params["include"] = include
        return await self.request(
            Route(method="GET", path=f"/reserved-slots/{slot_id}"),
            params=params,
        )

    async def list_reserved_slots(
        self,
        *,
        organization_id: int | None = None,
        player_id: int | None = None,
        server_id: int | None = None,
        search: str | None = None,
        expired: bool | None = None,
        include: str | None = None,
        page_size: int | None = None,
    ) -> dict[str, Any]:
        """List reserved slots."""
        params: dict[str, Any] = {}
        if organization_id:
            params["filter[organization]"] = organization_id
        if player_id:
            params["filter[player]"] = player_id
        if server_id:
            params["filter[server]"] = server_id
        if search:
            params["filter[search]"] = search
        if expired is not None:
            params["filter[expired]"] = str(expired).lower()
        if include:
            params["include"] = include
        if page_size:
            params["page[size]"] = page_size
        return await self.request(
            Route(method="GET", path="/reserved-slots"),
            params=params,
        )

    async def update_reserved_slot(
        self,
        slot_id: str,
        *,
        identifiers: list[str | dict[str, Any]] | None = None,
        expires: str | None = None,
        server_ids: list[int] | None = None,
    ) -> dict[str, Any]:
        """Update a reserved slot."""
        attributes = {"identifiers": identifiers or [], "expires": expires}
        relationships: dict[str, Any] = {}
        if server_ids is not None:
            relationships["servers"] = {
                "data": [{"type": "server", "id": str(sid)} for sid in server_ids],
            }
        data = {
            "data": {
                "type": "reservedSlot",
                "id": slot_id,
                "attributes": attributes,
                "relationships": relationships,
            },
        }
        return await self.request(
            Route(method="PATCH", path=f"/reserved-slots/{slot_id}"),
            json=data,
        )

    async def delete_reserved_slot(self, slot_id: str) -> None:
        """Delete a reserved slot."""
        await self.request(Route(method="DELETE", path=f"/reserved-slots/{slot_id}"))

    # -------------------------------- Servers ----------------------------- #

    async def list_servers(
        self,
        *,
        search: str | None = None,
        game: str | None = None,
        status: str | None = None,
        countries: list[str] | None = None,
        page_size: int | None = None,
        sort: str | None = None,
    ) -> dict[str, Any]:
        """List/search servers."""
        params: dict[str, Any] = {}
        if search:
            params["filter[search]"] = search
        if game:
            params["filter[game]"] = game
        if status:
            params["filter[status]"] = status
        if countries:
            # API accepts repeated countries or comma-separated; we use comma-separated list
            params["filter[countries]"] = ",".join(countries)
        if page_size:
            params["page[size]"] = page_size
        if sort:
            params["sort"] = sort
        return await self.request(Route(method="GET", path="/servers"), params=params)

    async def get_server(
        self,
        server_id: int,
        *,
        include: str | None = None,
    ) -> dict[str, Any]:
        """Get server information."""
        params: dict[str, Any] = {}
        if include:
            params["include"] = include
        return await self.request(
            Route(method="GET", path=f"/servers/{server_id}"),
            params=params,
        )

    async def server_player_count_history(
        self,
        server_id: int,
        *,
        start: str,
        stop: str,
        resolution: Literal["raw", "30", "60", "1440"] | None = None,
    ) -> dict[str, Any]:
        """Get server player count history."""
        params: dict[str, Any] = {"start": start, "stop": stop}
        if resolution:
            params["resolution"] = resolution
        return await self.request(
            Route(method="GET", path=f"/servers/{server_id}/player-count-history"),
            params=params,
        )

    async def server_rank_history(
        self,
        server_id: int,
        *,
        start: str,
        stop: str,
    ) -> dict[str, Any]:
        """Get server rank history."""
        params = {"start": start, "stop": stop}
        return await self.request(
            Route(method="GET", path=f"/servers/{server_id}/rank-history"),
            params=params,
        )

    async def server_group_rank_history(
        self,
        server_id: int,
        *,
        start: str,
        stop: str,
    ) -> dict[str, Any]:
        """Get server group rank history."""
        params = {"start": start, "stop": stop}
        return await self.request(
            Route(method="GET", path=f"/servers/{server_id}/group-rank-history"),
            params=params,
        )

    async def server_time_played_history(
        self,
        server_id: int,
        *,
        start: str,
        stop: str,
    ) -> dict[str, Any]:
        """Get server time played history."""
        params = {"start": start, "stop": stop}
        return await self.request(
            Route(method="GET", path=f"/servers/{server_id}/time-played-history"),
            params=params,
        )

    async def server_unique_player_history(
        self,
        server_id: int,
        *,
        start: str,
        stop: str,
    ) -> dict[str, Any]:
        """Get server unique player history."""
        params = {"start": start, "stop": stop}
        return await self.request(
            Route(method="GET", path=f"/servers/{server_id}/unique-player-history"),
            params=params,
        )

    async def server_sessions(
        self,
        server_id: int,
        *,
        start: str | None = None,
        stop: str | None = None,
        at: str | None = None,
        include: str | None = None,
    ) -> dict[str, Any]:
        """List sessions for a server (relationships endpoint)."""
        params: dict[str, Any] = {}
        if start:
            params["start"] = start
        if stop:
            params["stop"] = stop
        if at:
            params["at"] = at
        if include:
            params["include"] = include
        return await self.request(
            Route(method="GET", path=f"/servers/{server_id}/relationships/sessions"),
            params=params,
        )

    async def server_outages(
        self,
        server_id: int,
        *,
        range_: str | None = None,
        include: str | None = None,
        page_size: int | None = None,
    ) -> dict[str, Any]:
        """Get server outage history."""
        params: dict[str, Any] = {}
        if range_:
            params["filter[range]"] = range_
        if include:
            params["include"] = include
        if page_size:
            params["page[size]"] = page_size
        return await self.request(
            Route(method="GET", path=f"/servers/{server_id}/relationships/outages"),
            params=params,
        )

    async def server_downtime(
        self,
        server_id: int,
        *,
        start: str,
        stop: str,
        include: str | None = None,
        page_size: int | None = None,
    ) -> dict[str, Any]:
        """Get server downtime history."""
        params: dict[str, Any] = {"start": start, "stop": stop}
        if include:
            params["include"] = include
        if page_size:
            params["page[size]"] = page_size
        return await self.request(
            Route(method="GET", path=f"/servers/{server_id}/relationships/downtime"),
            params=params,
        )

    async def server_first_time_history(
        self,
        server_id: int,
        *,
        start: str,
        stop: str,
    ) -> dict[str, Any]:
        """Get server first-time player history."""
        params = {"start": start, "stop": stop}
        return await self.request(
            Route(method="GET", path=f"/servers/{server_id}/first-time-history"),
            params=params,
        )

    async def create_server(
        self,
        *,
        game: str,
        ip: str,
        port: int,
        port_query: int,
        organization_id: int,
    ) -> dict[str, Any]:
        """Create a server."""
        data = {
            "data": {
                "type": "server",
                "attributes": {
                    "ip": ip,
                    "port": port,
                    "portQuery": port_query,
                },
                "relationships": {
                    "game": {"data": {"type": "game", "id": game}},
                    "organization": {
                        "data": {"type": "organization", "id": str(organization_id)},
                    },
                },
            },
        }
        return await self.request(Route(method="POST", path="/servers"), json=data)

    async def update_server(
        self,
        server_id: int,
        *,
        metadata: dict[str, Any] | None = None,
        port_query: int | None = None,
        default_ban_list_id: str | None = None,
        server_group_id: str | None = None,
    ) -> dict[str, Any]:
        """Update a server."""
        attributes: dict[str, Any] = {}
        if metadata is not None:
            attributes["metadata"] = metadata
        if port_query is not None:
            attributes["portQuery"] = port_query
        relationships: dict[str, Any] = {}
        if default_ban_list_id is not None:
            relationships["defaultBanList"] = {
                "data": {"type": "banList", "id": default_ban_list_id},
            }
        if server_group_id is not None:
            relationships["serverGroup"] = {
                "data": {"type": "serverGroup", "id": server_group_id},
            }
        data: dict[str, Any] = {
            "data": {
                "type": "server",
                "id": str(server_id),
                "attributes": attributes,
            },
        }
        if relationships:
            data["data"]["relationships"] = relationships
        return await self.request(
            Route(method="PATCH", path=f"/servers/{server_id}"),
            json=data,
        )

    async def enable_server_rcon(
        self,
        server_id: int,
        *,
        password: str,
        port: int | None = None,
        ip: str | None = None,
    ) -> dict[str, Any]:
        """Enable RCON for a server."""
        attributes: dict[str, Any] = {"password": password}
        if port is not None:
            attributes["port"] = port
        if ip is not None:
            attributes["ip"] = ip
        data = {
            "data": {
                "type": "server",
                "attributes": {"rcon": attributes},
            },
        }
        return await self.request(
            Route(method="POST", path=f"/servers/{server_id}/rcon"),
            json=data,
        )

    async def delete_server_rcon(self, server_id: int) -> None:
        """Delete RCON configuration for a server."""
        await self.request(Route(method="DELETE", path=f"/servers/{server_id}/rcon"))

    async def disconnect_server_rcon(self, server_id: int) -> None:
        """Disconnect RCON for a server."""
        await self.request(
            Route(method="DELETE", path=f"/servers/{server_id}/rcon/disconnect"),
        )

    async def connect_server_rcon(self, server_id: int) -> None:
        """Connect RCON for a server."""
        await self.request(
            Route(method="DELETE", path=f"/servers/{server_id}/rcon/connect"),
        )

    async def force_update_server(self, server_id: int) -> dict[str, Any]:
        """Force update a server."""
        return await self.request(
            Route(method="POST", path=f"/servers/{server_id}/force-update"),
        )

    # -------------------------------- Sessions ---------------------------- #

    async def list_sessions(
        self,
        *,
        servers: list[int] | None = None,
        games: list[str] | None = None,
        organizations: list[int] | None = None,
        players: list[int] | None = None,
        identifiers: list[str] | None = None,
        range_: str | None = None,
        at: str | None = None,
        include: str | None = None,
        page_size: int | None = None,
    ) -> dict[str, Any]:
        """List sessions."""
        params: dict[str, Any] = {}
        if servers:
            params["filter[servers]"] = ",".join(map(str, servers))
        if games:
            params["filter[games]"] = ",".join(games)
        if organizations:
            params["filter[organizations]"] = ",".join(map(str, organizations))
        if players:
            params["filter[players]"] = ",".join(map(str, players))
        if identifiers:
            params["filter[identifiers]"] = ",".join(identifiers)
        if range_:
            params["filter[range]"] = range_
        if at:
            params["filter[at]"] = at
        if include:
            params["include"] = include
        if page_size:
            params["page[size]"] = page_size
        return await self.request(Route(method="GET", path="/sessions"), params=params)

    async def get_session(
        self,
        session_id: str,
        *,
        include: str | None = None,
    ) -> dict[str, Any]:
        """Get a specific session."""
        params: dict[str, Any] = {}
        if include:
            params["include"] = include
        return await self.request(
            Route(method="GET", path=f"/sessions/{session_id}"),
            params=params,
        )

    async def session_coplay(
        self,
        session_id: str,
        *,
        page_size: int | None = None,
        include: str | None = None,
    ) -> dict[str, Any]:
        """Get coplay data for a session."""
        params: dict[str, Any] = {}
        if page_size:
            params["page[size]"] = page_size
        if include:
            params["include"] = include
        return await self.request(
            Route(method="GET", path=f"/sessions/{session_id}/relationships/coplay"),
            params=params,
        )

    # --------------------------------- Users ------------------------------ #

    async def get_user(self, user_id: int) -> dict[str, Any]:
        """Get user information."""
        return await self.request(Route(method="GET", path=f"/users/{user_id}"))

    # -------------------------------- Player Notes ----------------------- #

    async def list_player_notes(
        self,
        player_id: int,
        *,
        expired: bool | None = None,
        organizations: list[int] | None = None,
        personal: bool | None = None,
        search: str | None = None,
        users: list[int] | None = None,
        include: str | None = None,
        page_size: int | None = None,
        page_key: str | None = None,
        page_rel: Literal["next", "prev"] | None = None,
    ) -> dict[str, Any]:
        """List player notes for a player."""
        params: dict[str, Any] = {}
        if expired is not None:
            params["filter[expired]"] = str(expired).lower()
        if organizations:
            params["filter[organizations]"] = ",".join(map(str, organizations))
        if personal is not None:
            params["filter[personal]"] = str(personal).lower()
        if search:
            params["filter[search]"] = search
        if users:
            params["filter[users]"] = ",".join(map(str, users))
        if include:
            params["include"] = include
        if page_size:
            params["page[size]"] = page_size
        if page_key:
            params["page[key]"] = page_key
        if page_rel:
            params["page[rel]"] = page_rel
        return await self.request(
            Route(method="GET", path=f"/players/{player_id}/relationships/notes"),
            params=params,
        )

    async def get_player_note(self, player_id: int, note_id: str) -> dict[str, Any]:
        """Get a specific player note."""
        return await self.request(
            Route(
                method="GET",
                path=f"/players/{player_id}/relationships/notes/{note_id}",
            ),
        )

    async def create_player_note(
        self,
        player_id: int,
        *,
        note: str,
        shared: bool,
        expires_at: str | None = None,
        clearance_level: int | None = None,
        organization_id: int | None = None,
        trigger_id: str | None = None,
    ) -> dict[str, Any]:
        """Create a player note."""
        relationships: dict[str, Any] = {
            "player": {"data": {"type": "player", "id": str(player_id)}},
        }
        if organization_id is not None:
            relationships["organization"] = {
                "data": {"type": "organization", "id": str(organization_id)},
            }
        if trigger_id is not None:
            relationships["trigger"] = {
                "data": {"type": "trigger", "id": trigger_id},
            }
        data = {
            "data": {
                "type": "playerNote",
                "attributes": {
                    "note": note,
                    "shared": shared,
                    "expiresAt": expires_at,
                    "clearanceLevel": clearance_level,
                },
                "relationships": relationships,
            },
        }
        return await self.request(
            Route(method="POST", path=f"/players/{player_id}/relationships/notes"),
            json=data,
        )

    async def update_player_note(
        self,
        player_id: int,
        note_id: str,
        *,
        note: str | None = None,
        shared: bool | None = None,
        expires_at: str | None = None,
        clearance_level: int | None = None,
    ) -> dict[str, Any]:
        """Update a player note."""
        attributes = {
            "note": note,
            "shared": shared,
            "expiresAt": expires_at,
            "clearanceLevel": clearance_level,
        }
        attributes = {k: v for k, v in attributes.items() if v is not None}
        data = {
            "data": {
                "type": "playerNote",
                "id": note_id,
                "attributes": attributes,
            },
        }
        return await self.request(
            Route(
                method="PATCH",
                path=f"/players/{player_id}/relationships/notes/{note_id}",
            ),
            json=data,
        )

    async def delete_player_note(self, player_id: int, note_id: str) -> None:
        """Delete a player note."""
        await self.request(
            Route(
                method="DELETE",
                path=f"/players/{player_id}/relationships/notes/{note_id}",
            ),
        )

    # -------------------------- Organization Friends ----------------------- #

    async def list_organization_friends(
        self,
        organization_id: int,
        *,
        include: str | None = None,
        page_size: int | None = None,
    ) -> dict[str, Any]:
        """List organization friends."""
        params: dict[str, Any] = {}
        if include:
            params["include"] = include
        if page_size:
            params["page[size]"] = page_size
        return await self.request(
            Route(
                method="GET",
                path=f"/organizations/{organization_id}/relationships/friends",
            ),
            params=params,
        )

    async def get_organization_friend(
        self,
        organization_id: int,
        friend_id: int,
        *,
        include: str | None = None,
    ) -> dict[str, Any]:
        """Get a specific organization friend."""
        params: dict[str, Any] = {}
        if include:
            params["include"] = include
        return await self.request(
            Route(
                method="GET",
                path=f"/organizations/{organization_id}/relationships/friends/{friend_id}",
            ),
            params=params,
        )

    async def create_organization_friend(
        self,
        organization_id: int,
        friend_id: int,
        *,
        identifiers: list[str] | None = None,
        notes: bool = False,
        player_flags: list[str] | None = None,
    ) -> dict[str, Any]:
        """Create an organization friend request."""
        relationships: dict[str, Any] = {
            "friend": {"data": {"type": "organization", "id": str(friend_id)}},
        }
        if player_flags:
            relationships["playerFlags"] = {
                "data": [
                    {"type": "playerFlag", "id": flag_id} for flag_id in player_flags
                ],
            }
        data = {
            "data": {
                "type": "organizationFriend",
                "attributes": {
                    "identifiers": identifiers or [],
                    "notes": notes,
                },
                "relationships": relationships,
            },
        }
        return await self.request(
            Route(
                method="POST",
                path=f"/organizations/{organization_id}/relationships/friends",
            ),
            json=data,
        )

    async def update_organization_friend(
        self,
        organization_id: int,
        friend_id: int,
        *,
        identifiers: list[str] | None = None,
        notes: bool | None = None,
        player_flags: list[str] | None = None,
    ) -> dict[str, Any]:
        """Update an organization friend."""
        attributes: dict[str, Any] = {}
        if identifiers is not None:
            attributes["identifiers"] = identifiers
        if notes is not None:
            attributes["notes"] = notes
        relationships: dict[str, Any] = {}
        if player_flags is not None:
            relationships["playerFlags"] = {
                "data": [
                    {"type": "playerFlag", "id": flag_id} for flag_id in player_flags
                ],
            }
        data: dict[str, Any] = {
            "data": {
                "type": "organizationFriend",
                "attributes": attributes,
            },
        }
        if relationships:
            data["data"]["relationships"] = relationships
        return await self.request(
            Route(
                method="PATCH",
                path=f"/organizations/{organization_id}/relationships/friends/{friend_id}",
            ),
            json=data,
        )

    async def delete_organization_friend(
        self,
        organization_id: int,
        friend_id: int,
    ) -> None:
        """Delete an organization friend."""
        await self.request(
            Route(
                method="DELETE",
                path=f"/organizations/{organization_id}/relationships/friends/{friend_id}",
            ),
        )

    async def bulk_delete_organization_friends(
        self,
        organization_id: int,
        friend_ids: list[int],
    ) -> None:
        """Bulk delete organization friends."""
        data = {
            "data": [{"type": "organization", "id": str(fid)} for fid in friend_ids],
        }
        await self.request(
            Route(
                method="DELETE",
                path=f"/organizations/{organization_id}/relationships/friends",
            ),
            json=data,
        )
